"""
Implementation of job for the `asyncio` scheduler.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

from __future__ import annotations

from logging import Logger
from typing import Any, Callable, Coroutine

from scheduler.base.job import BaseJob


class Job(BaseJob[Callable[..., Coroutine[Any, Any, None]]]):
    r"""
    |AioJob| class bundling time and callback function methods.

    Parameters
    ----------
    job_type : JobType
        Indicator which defines which calculations has to be used.
    timing : TimingWeekly
        Desired execution time(s).
    handle : Callable[..., None]
        Handle to a callback function.
    args : tuple[Any]
        Positional argument payload for the function handle within a |AioJob|.
    kwargs : Optional[dict[str, Any]]
        Keyword arguments payload for the function handle within a |AioJob|.
    max_attempts : Optional[int]
        Number of times the |AioJob| will be executed where ``0 <=> inf``.
        A |AioJob| with no free attempt will be deleted.
    tags : Optional[set[str]]
        The tags of the |AioJob|.
    delay : Optional[bool]
        *Deprecated*: If ``True`` wait with the execution for the next scheduled time.
    start : Optional[datetime.datetime]
        Set the reference `datetime.datetime` stamp the |AioJob|
        will be scheduled against. Default value is `datetime.datetime.now()`.
    stop : Optional[datetime.datetime]
        Define a point in time after which a |AioJob| will be stopped
        and deleted.
    skip_missing : Optional[bool]
        If ``True`` a |AioJob| will only schedule it's newest planned
        execution and drop older ones.
    alias : Optional[str]
        Overwrites the function handle name in the string representation.
    tzinfo : Optional[datetime.tzinfo]
        Set the timezone of the |AioScheduler| the |AioJob|
        is scheduled in.

    Returns
    -------
    Job
        Instance of a scheduled |AioJob|.
    """

    # pylint: disable=no-member invalid-name

    async def _exec(self, logger: Logger) -> None:
        coroutine = self._BaseJob__handle(*self._BaseJob__args, **self._BaseJob__kwargs)  # type: ignore
        try:
            await coroutine
        except Exception:
            logger.exception("Unhandled exception in `%r`!", self)
            self._BaseJob__failed_attempts += 1  # type: ignore
        self._BaseJob__attempts += 1  # type: ignore

    # pylint: enable=no-member invalid-name

    def __repr__(self) -> str:
        params: tuple[str, ...] = self._repr()
        return f"scheduler.asyncio.job.Job({', '.join(params)})"



"""
Implementation of a `threading` compatible in-process scheduler.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

from scheduler.error import SchedulerError
from scheduler.threading.scheduler import Scheduler

__all__ = ["SchedulerError", "Scheduler"]



"""
Defines the typing for all trigger objects.

Combines custom trigger objects like Weekday with the python in build
types for the datetime library.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

import datetime as dt
from typing import Union

from scheduler.trigger.core import Weekday

# execution interval
TimingCyclic = dt.timedelta  # Scheduler
_TimingCyclicList = list[TimingCyclic]

# time on the clock
_TimingDaily = dt.time  # JobTimer
# TimingDaily = Union[dt.time, list[dt.time]]
_TimingDailyList = list[_TimingDaily]  # Job
TimingDailyUnion = Union[_TimingDaily, _TimingDailyList]  # Scheduler

# day of the week or time on the clock
_TimingWeekly = Weekday
_TimingWeeklyList = list[_TimingWeekly]
TimingWeeklyUnion = Union[_TimingWeekly, _TimingWeeklyList]  # Scheduler

TimingJobTimerUnion = Union[TimingCyclic, _TimingDaily, _TimingWeekly]  # JobTimer
TimingJobUnion = Union[_TimingCyclicList, _TimingDailyList, _TimingWeeklyList]  # Job

TimingOnceUnion = Union[dt.datetime, TimingCyclic, _TimingWeekly, _TimingDaily]  # Scheduler.once



"""
Information and error messages.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

DUPLICATE_EFFECTIVE_TIME = "Times that are effectively identical are not allowed."

CYCLIC_TYPE_ERROR_MSG = "Wrong input for Cyclic! Expected input type:\n" + "datetime.timedelta"
_DAILY_TYPE_ERROR_MSG = (
    "Wrong input for {0}! Select one of the following input types:\n"
    + "datetime.time | list[datetime.time]"
)
MINUTELY_TYPE_ERROR_MSG = _DAILY_TYPE_ERROR_MSG.format("Minutely")
HOURLY_TYPE_ERROR_MSG = _DAILY_TYPE_ERROR_MSG.format("Hourly")
DAILY_TYPE_ERROR_MSG = _DAILY_TYPE_ERROR_MSG.format("Daily")
WEEKLY_TYPE_ERROR_MSG = (
    "Wrong input for Weekly! Select one of the following input types:\n"
    + "DAY | list[DAY]\n"
    + "where `DAY = Weekday`"
)

TZ_ERROR_MSG = "Can't use offset-naive and offset-aware datetimes together."
_TZ_ERROR_MSG = TZ_ERROR_MSG[:-1] + " for {0}."

START_STOP_ERROR = "Start argument must be smaller than the stop argument."

ONCE_TYPE_ERROR_MSG = (
    "Wrong input for Once! Select one of the following input types:\n"
    + "dt.datetime | dt.timedelta | Weekday | dt.time"
)



"""
Implementation of a `BaseScheduler`.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterable
from functools import wraps
from logging import Logger, getLogger
from typing import Any, Callable, Generic, List, Optional, TypeVar

from scheduler.base.job import BaseJobType
from scheduler.base.timingtype import (
    TimingCyclic,
    TimingDailyUnion,
    TimingOnceUnion,
    TimingWeeklyUnion,
)

# TODO:
# import sys
# if sys.version_info < (3, 10):
#     from typing_extensions import ParamSpec
# else:
#     from typing import ParamSpec


LOGGER = getLogger("scheduler")


def select_jobs_by_tag(
    jobs: set[BaseJobType],
    tags: set[str],
    any_tag: bool,
) -> set[BaseJobType]:
    r"""
    Select |BaseJob|\ s by matching `tags`.

    Parameters
    ----------
    jobs : set[BaseJob]
        Unfiltered set of |BaseJob|\ s.
    tags : set[str]
        Tags to filter |BaseJob|\ s.
    any_tag : bool
        False: To match a |BaseJob| all tags have to match.
        True: To match a |BaseJob| at least one tag has to match.

    Returns
    -------
    set[BaseJob]
        Selected |BaseJob|\ s.
    """
    if any_tag:
        return {job for job in jobs if tags & job.tags}
    return {job for job in jobs if tags <= job.tags}


def deprecated(fields: List[str]) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Decorator for marking specified function arguments as deprecated.

    Parameters
    ----------
    fields : List[str]
        A list of strings representing the names of the function arguments that are deprecated.

    Examples
    --------
    @deprecated(['old_arg'])
    def some_function(new_arg, old_arg=None):
        pass

    Calling `some_function(new_arg=5, old_arg=3)` generates a deprecation warning for using 'old_arg'.
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def real_wrapper(*args: tuple[Any, ...], **kwargs: dict[str, Any]) -> Any:
            for f in fields:
                if f in kwargs and kwargs[f] is not None:
                    # keep it in kwargs
                    warnings.warn(
                        (
                            f"Using the `{f}` argument is deprecated and will "
                            "be removed in the next minor release."
                        ),
                        DeprecationWarning,
                        stacklevel=3,
                    )
            return func(*args, **kwargs)

        return real_wrapper

    return wrapper


T = TypeVar("T", bound=Callable[[], Any])


class BaseScheduler(
    ABC, Generic[BaseJobType, T]
):  # NOTE maybe a typing Protocol class is better than an ABC class
    """
    Interface definition of an abstract scheduler.

    Author: Jendrik A. Potyka, Fabian A. Preiss
    """

    _logger: Logger

    def __init__(self, logger: Optional[Logger] = None) -> None:
        self._logger = logger if logger else LOGGER

    @abstractmethod
    def delete_job(self, job: BaseJobType) -> None:
        """Delete a |BaseJob| from the `BaseScheduler`."""

    @abstractmethod
    def delete_jobs(
        self,
        tags: Optional[set[str]] = None,
        any_tag: bool = False,
    ) -> int:
        r"""Delete a set of |BaseJob|\ s from the `BaseScheduler` by tags."""

    @abstractmethod
    def get_jobs(
        self,
        tags: Optional[set[str]] = None,
        any_tag: bool = False,
    ) -> set[BaseJobType]:
        r"""Get a set of |BaseJob|\ s from the `BaseScheduler` by tags."""

    @abstractmethod
    def cyclic(self, timing: TimingCyclic, handle: T, **kwargs) -> BaseJobType:
        """Schedule a cyclic |BaseJob|."""

    @abstractmethod
    def minutely(self, timing: TimingDailyUnion, handle: T, **kwargs) -> BaseJobType:
        """Schedule a minutely |BaseJob|."""

    @abstractmethod
    def hourly(self, timing: TimingDailyUnion, handle: T, **kwargs) -> BaseJobType:
        """Schedule an hourly |BaseJob|."""

    @abstractmethod
    def daily(self, timing: TimingDailyUnion, handle: T, **kwargs) -> BaseJobType:
        """Schedule a daily |BaseJob|."""

    @abstractmethod
    def weekly(self, timing: TimingWeeklyUnion, handle: T, **kwargs) -> BaseJobType:
        """Schedule a weekly |BaseJob|."""

    @abstractmethod
    def once(
        self,
        timing: TimingOnceUnion,
        handle: T,
        *,
        args: Optional[tuple[Any]] = None,
        kwargs: Optional[dict[str, Any]] = None,
        tags: Optional[Iterable[str]] = None,
        alias: Optional[str] = None,
    ) -> BaseJobType:
        """Schedule a oneshot |BaseJob|."""

    @property
    @abstractmethod
    def jobs(self) -> set[BaseJobType]:
        r"""Get the set of all |BaseJob|\ s."""



"""
Implementation of essential functions and components for a `BaseJob`.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

import datetime as dt
from typing import Optional, Union, cast

from scheduler.base.job import BaseJobType
from scheduler.base.timingtype import (
    TimingCyclic,
    TimingDailyUnion,
    TimingJobUnion,
    TimingWeeklyUnion,
)
from scheduler.error import SchedulerError


def str_cutoff(string: str, max_length: int, cut_tail: bool = False) -> str:
    """
    Abbreviate a string to a given length.

    The resulting string will carry an indicator if it's abbreviated,
    like ``stri#``.

    Parameters
    ----------
    string : str
        String which is to be cut.
    max_length : int
        Max resulting string length.
    cut_tail : bool
        ``False`` for string abbreviation from the front, else ``True``.

    Returns
    -------
    str
        Resulting string
    """
    if max_length < 1:
        raise ValueError("max_length < 1 not allowed")

    if len(string) > max_length:
        pos = max_length - 1
        return string[:pos] + "#" if cut_tail else "#" + string[-pos:]

    return string


def check_tzname(tzinfo: Optional[dt.tzinfo]) -> Optional[str]:
    """Composed of the datetime.datetime.tzname and the datetime._check_tzname methode."""
    if tzinfo is None:
        return None
    name: Optional[str] = tzinfo.tzname(None)
    if not isinstance(name, str):
        raise SchedulerError(f"tzinfo.tzname() must return None or string, not {type(name)}")
    return name


def create_job_instance(
    job_class: type[BaseJobType],
    timing: Union[TimingCyclic, TimingDailyUnion, TimingWeeklyUnion],
    **kwargs,
) -> BaseJobType:
    """Create a job instance from the given input parameters."""
    if not isinstance(timing, list):
        timing_list = cast(TimingJobUnion, [timing])
    else:
        timing_list = cast(TimingJobUnion, timing)

    return job_class(
        timing=timing_list,
        **kwargs,
    )



import datetime as dt
import doctest
from typing import Any

import pytest

from .helpers import T_2021_5_26__3_55


# NOTE: We cannot test for the full table, as some Jobs depend on the time of execution
#       e.g. a Job supposed to run on Weekday.MONDAY. The ordering between the Jobs scheduled
#       at 0:09:59 can be guaranteed though, as they differ on the milliseconds level.
# NOTE: Currently when updating the example in the README.md file, the changes should be applied
#       manually in this file as well.
# NOTE: The same example and doctest can be found in `doc/examples/general_job_scheduling.rst`,
#       however here the test is more granular, wheras in `doc/examples` the focus is more on
#       readability and additional comments.
@pytest.mark.parametrize(
    "patch_datetime_now",
    [[T_2021_5_26__3_55 + dt.timedelta(microseconds=x) for x in range(17)]],
    indirect=["patch_datetime_now"],
)
def test_general_readme(patch_datetime_now: Any) -> None:
    r"""
    >>> import datetime as dt
    >>> from scheduler import Scheduler
    >>> from scheduler.trigger import Monday, Tuesday

    >>> def foo():
    ...     print("foo")

    >>> schedule = Scheduler()

    >>> schedule.cyclic(dt.timedelta(minutes=10), foo)  # doctest:+ELLIPSIS
    scheduler.Job(<JobType.CYCLIC...>, [datetime.timedelta(seconds=600)], <function foo at 0x...>, (), {}, 0, 1, True, datetime.datetime(...), None, False, None, None)

    >>> schedule.minutely(dt.time(second=15), foo)  # doctest:+ELLIPSIS
    scheduler.Job(<JobType.MINUTELY...>, [datetime.time(0, 0, 15)], <function foo at 0x...>, (), {}, 0, 1, True, datetime.datetime(...), None, False, None, None)

    >>> schedule.hourly(dt.time(minute=30, second=15), foo)  # doctest:+ELLIPSIS
    scheduler.Job(<JobType.HOURLY...>, [datetime.time(0, 30, 15)], <function foo at 0x...>, (), {}, 0, 1, True, datetime.datetime(...), None, False, None, None)

    >>> schedule.daily(dt.time(hour=16, minute=30), foo)  # doctest:+ELLIPSIS
    scheduler.Job(<JobType.DAILY...>, [datetime.time(16, 30)], <function foo at 0x...>, (), {}, 0, 1, True, datetime.datetime(...), None, False, None, None)

    >>> schedule.weekly(Monday(), foo)  # doctest:+ELLIPSIS
    scheduler.Job(<JobType.WEEKLY...>, [Monday(time=datetime.time(0, 0))], <function foo at 0x...>, (), {}, 0, 1, True, datetime.datetime(...), None, False, None, None)

    >>> schedule.weekly(Monday(dt.time(hour=16, minute=30)), foo)  # doctest:+ELLIPSIS
    scheduler.Job(<JobType.WEEKLY...>, [Monday(time=datetime.time(16, 30))], <function foo at 0x...>, (), {}, 0, 1, True, datetime.datetime(...), None, False, None, None)

    >>> schedule.once(dt.timedelta(minutes=10), foo)  # doctest:+ELLIPSIS
    scheduler.Job(<JobType.CYCLIC...>, [datetime.timedelta(seconds=600)], <function foo at 0x...>, (), {}, 1, 1, True, datetime.datetime(...), None, False, None, None)

    >>> schedule.once(Tuesday(), foo)  # doctest:+ELLIPSIS
    scheduler.Job(<JobType.WEEKLY...>, [Tuesday(time=datetime.time(0, 0))], <function foo at 0x...>, (), {}, 1, 1, True, datetime.datetime(...), None, False, None, None)

    >>> schedule.once(dt.datetime(year=2022, month=2, day=15, minute=45), foo)  # doctest:+ELLIPSIS
    scheduler.Job(<JobType.CYCLIC...>, [datetime.timedelta(0)], <function foo at 0x...>, (), {}, 1, 1, False, datetime.datetime(2022, 2, 15, 0, 45), None, False, None, None)


    >>> print(schedule)  # doctest:+ELLIPSIS
    max_exec=inf, tzinfo=None, priority_function=linear_priority_function, #jobs=9
    <BLANKLINE>
    type     function / alias due at                 due in      attempts weight
    -------- ---------------- ------------------- --------- ------------- ------
    MINUTELY foo()            2021-05-26 03:55:15   0:00:14         0/inf      1
    CYCLIC   foo()            2021-05-26 04:05:00   0:09:59         0/inf      1
    ONCE     foo()            2021-05-26 04:05:00   0:09:59           0/1      1
    HOURLY   foo()            2021-05-26 04:30:15   0:35:14         0/inf      1
    DAILY    foo()            2021-05-26 16:30:00  12:34:59         0/inf      1
    WEEKLY   foo()            2021-05-31 00:00:00    4 days         0/inf      1
    WEEKLY   foo()            2021-05-31 16:30:00    5 days         0/inf      1
    ONCE     foo()            2021-06-01 00:00:00    5 days           0/1      1
    ONCE     foo()            2022-02-15 00:45:00  264 days           0/1      1
    <BLANKLINE>


    >>> import time
    >>> while True:  # doctest:+SKIP
    ...     schedule.exec_jobs()
    ...     time.sleep(1)
    """
    DP = doctest.DocTestParser()
    assert test_general_readme.__doc__
    dt_readme = DP.get_doctest(test_general_readme.__doc__, globals(), "README", None, None)
    DTR = doctest.DocTestRunner()
    assert doctest.TestResults(failed=0, attempted=16) == DTR.run(dt_readme)



"""
A simple in-process python `scheduler` library for `datetime` objects.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

__version__ = "0.8.7"
__author__ = "Jendrik A. Potyka, Fabian A. Preiss"

from scheduler.error import SchedulerError
from scheduler.threading.scheduler import Scheduler

__all__ = ["SchedulerError", "Scheduler"]



"""
Implementation of essential functions for a `BaseJob`.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

from __future__ import annotations

import datetime as dt
from typing import Optional, cast

import typeguard as tg

from scheduler.base.definition import JOB_TIMING_TYPE_MAPPING, JobType
from scheduler.base.job_timer import JobTimer
from scheduler.base.timingtype import TimingJobUnion
from scheduler.error import SchedulerError
from scheduler.message import (
    _TZ_ERROR_MSG,
    DUPLICATE_EFFECTIVE_TIME,
    START_STOP_ERROR,
    TZ_ERROR_MSG,
)
from scheduler.trigger.core import Weekday
from scheduler.util import are_times_unique, are_weekday_times_unique


def prettify_timedelta(timedelta: dt.timedelta) -> str:
    """
    Humanize timedelta string readability for negative values.

    Parameters
    ----------
    timedelta : datetime.timedelta
        datetime instance

    Returns
    -------
    str
        Human readable string representation rounded to seconds
    """
    seconds = timedelta.total_seconds()
    if seconds < 0:
        res = f"-{-timedelta}"
    else:
        res = str(timedelta)
    return res.split(",")[0].split(".")[0]


def get_pending_timer(timers: list[JobTimer]) -> JobTimer:
    """Get the the timer with the largest overdue time."""
    unsorted_timer_datetimes: dict[JobTimer, dt.datetime] = {}
    for timer in timers:
        unsorted_timer_datetimes[timer] = timer.datetime
    sorted_timers = sorted(
        unsorted_timer_datetimes,
        key=unsorted_timer_datetimes.get,  # type: ignore
    )
    return sorted_timers[0]


def sane_timing_types(job_type: JobType, timing: TimingJobUnion) -> None:
    """
    Determine if the `JobType` is fulfilled by the type of the specified `timing`.

    Parameters
    ----------
    job_type : JobType
        :class:`~scheduler.job.JobType` to test agains.
    timing : TimingJobUnion
        The `timing` object to be tested.

    Raises
    ------
    TypeError
        If the `timing` object has the wrong `Type` for a specific `JobType`.
    """
    try:
        tg.check_type(timing, JOB_TIMING_TYPE_MAPPING[job_type]["type"])
        if job_type == JobType.CYCLIC:
            if not len(timing) == 1:
                raise TypeError
    except TypeError as err:
        raise SchedulerError(JOB_TIMING_TYPE_MAPPING[job_type]["err"]) from err


def standardize_timing_format(job_type: JobType, timing: TimingJobUnion) -> TimingJobUnion:
    r"""
    Return timings in standardized form.

    Clears irrelevant time positionals for `JobType.MINUTELY` and `JobType.HOURLY`.
    """
    if job_type is JobType.MINUTELY:
        timing = [time.replace(hour=0, minute=0) for time in cast(list[dt.time], timing)]
    elif job_type is JobType.HOURLY:
        timing = [time.replace(hour=0) for time in cast(list[dt.time], timing)]
    return timing


def check_timing_tzinfo(
    job_type: JobType,
    timing: TimingJobUnion,
    tzinfo: Optional[dt.tzinfo],
) -> None:
    """Raise if `timing` incompatible with `tzinfo` for `job_type`."""
    if job_type is JobType.WEEKLY:
        for weekday in cast(list[Weekday], timing):
            if bool(weekday.time.tzinfo) ^ bool(tzinfo):
                raise SchedulerError(TZ_ERROR_MSG)
    elif job_type in (JobType.MINUTELY, JobType.HOURLY, JobType.DAILY):
        for time in cast(list[dt.time], timing):
            if bool(time.tzinfo) ^ bool(tzinfo):
                raise SchedulerError(TZ_ERROR_MSG)


def check_duplicate_effective_timings(
    job_type: JobType,
    timing: TimingJobUnion,
    tzinfo: Optional[dt.tzinfo],
) -> None:
    """Raise given timings are not effectively duplicates."""
    if job_type is JobType.WEEKLY:
        if not are_weekday_times_unique(cast(list[Weekday], timing), tzinfo):
            raise SchedulerError(DUPLICATE_EFFECTIVE_TIME)
    elif job_type in (
        JobType.MINUTELY,
        JobType.HOURLY,
        JobType.DAILY,
    ):
        if not are_times_unique(cast(list[dt.time], timing)):
            raise SchedulerError(DUPLICATE_EFFECTIVE_TIME)


def set_start_check_stop_tzinfo(
    start: Optional[dt.datetime],
    stop: Optional[dt.datetime],
    tzinfo: Optional[dt.tzinfo],
) -> dt.datetime:
    """Raise if `start`, `stop` and `tzinfo` incompatible; Make start."""
    if start:
        if bool(start.tzinfo) ^ bool(tzinfo):
            raise SchedulerError(_TZ_ERROR_MSG.format("start"))
    else:
        start = dt.datetime.now(tzinfo)
    if stop:
        if bool(stop.tzinfo) ^ bool(tzinfo):
            raise SchedulerError(_TZ_ERROR_MSG.format("stop"))
    if stop is not None:
        if start >= stop:
            raise SchedulerError(START_STOP_ERROR)
    return start



"""
Implementation of job for the `threading` scheduler.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

import datetime as dt
import threading
from logging import Logger
from typing import Any, Callable, Optional

from scheduler.base.definition import JobType
from scheduler.base.job import BaseJob
from scheduler.base.timingtype import TimingJobUnion


class Job(BaseJob[Callable[..., None]]):
    r"""
    |Job| class bundling time and callback function methods.

    Parameters
    ----------
    job_type : JobType
        Indicator which defines which calculations has to be used.
    timing : TimingWeekly
        Desired execution time(s).
    handle : Callable[..., None]
        Handle to a callback function.
    args : tuple[Any]
        Positional argument payload for the function handle within a |Job|.
    kwargs : Optional[dict[str, Any]]
        Keyword arguments payload for the function handle within a |Job|.
    max_attempts : Optional[int]
        Number of times the |Job| will be executed where ``0 <=> inf``.
        A |Job| with no free attempt will be deleted.
    tags : Optional[set[str]]
        The tags of the |Job|.
    delay : Optional[bool]
        *Deprecated*: If ``True`` wait with the execution for the next scheduled time.
    start : Optional[datetime.datetime]
        Set the reference `datetime.datetime` stamp the |Job|
        will be scheduled against. Default value is `datetime.datetime.now()`.
    stop : Optional[datetime.datetime]
        Define a point in time after which a |Job| will be stopped
        and deleted.
    skip_missing : Optional[bool]
        If ``True`` a |Job| will only schedule it's newest planned
        execution and drop older ones.
    alias : Optional[str]
        Overwrites the function handle name in the string representation.
    tzinfo : Optional[datetime.tzinfo]
        Set the timezone of the |Scheduler| the |Job|
        is scheduled in.
    weight : Optional[float]
        Relative `weight` against other |Job|\ s.

    Returns
    -------
    Job
        Instance of a scheduled |Job|.
    """

    __weight: float
    __lock: threading.RLock

    def __init__(
        self,
        job_type: JobType,
        timing: TimingJobUnion,
        handle: Callable[..., None],
        *,
        args: Optional[tuple[Any, ...]] = None,
        kwargs: Optional[dict[str, Any]] = None,
        max_attempts: int = 0,
        tags: Optional[set[str]] = None,
        delay: bool = True,
        start: Optional[dt.datetime] = None,
        stop: Optional[dt.datetime] = None,
        skip_missing: bool = False,
        alias: Optional[str] = None,
        tzinfo: Optional[dt.tzinfo] = None,
        weight: float = 1,
    ):
        super().__init__(
            job_type,
            timing,
            handle,
            args=args,
            kwargs=kwargs,
            max_attempts=max_attempts,
            tags=tags,
            delay=delay,
            start=start,
            stop=stop,
            skip_missing=skip_missing,
            alias=alias,
            tzinfo=tzinfo,
        )
        self.__lock = threading.RLock()
        self.__weight = weight

    # pylint: disable=no-member invalid-name

    def _exec(self, logger: Logger) -> None:
        """Execute the callback function."""
        with self.__lock:
            try:
                self._BaseJob__handle(*self._BaseJob__args, **self._BaseJob__kwargs)  # type: ignore
            except Exception:
                logger.exception("Unhandled exception in `%r`!", self)
                self._BaseJob__failed_attempts += 1  # type: ignore
            self._BaseJob__attempts += 1  # type: ignore

    # pylint: enable=no-member invalid-name

    def _calc_next_exec(self, ref_dt: dt.datetime) -> None:
        with self.__lock:
            super()._calc_next_exec(ref_dt)

    def __repr__(self) -> str:
        with self.__lock:
            params: tuple[str, ...] = self._repr()
            params_sum: str = ", ".join(params[:6] + (repr(self.__weight),) + params[6:])
            return f"scheduler.Job({params_sum})"

    def __str__(self) -> str:
        return f"{super().__str__()}, w={self.weight:.3g}"

    def timedelta(self, dt_stamp: Optional[dt.datetime] = None) -> dt.timedelta:
        with self.__lock:
            return super().timedelta(dt_stamp)

    @property
    def datetime(self) -> dt.datetime:
        with self.__lock:
            return super().datetime

    @property
    def weight(self) -> float:
        """
        Return the weight of the `Job` instance.

        Returns
        -------
        float
            |Job| `weight`.
        """
        return self.__weight

    @property
    def has_attempts_remaining(self) -> bool:
        with self.__lock:
            return super().has_attempts_remaining



"""
Implementation of a `BaseJob`.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

from __future__ import annotations

import datetime as dt
import warnings
from abc import ABC, abstractmethod
from logging import Logger
from typing import Any, Callable, Generic, Optional, TypeVar, cast

from scheduler.base.definition import JobType
from scheduler.base.job_timer import JobTimer
from scheduler.base.job_util import (
    check_duplicate_effective_timings,
    check_timing_tzinfo,
    get_pending_timer,
    prettify_timedelta,
    sane_timing_types,
    set_start_check_stop_tzinfo,
    standardize_timing_format,
)
from scheduler.base.timingtype import TimingJobUnion

T = TypeVar("T", bound=Callable[[], Any])


class BaseJob(ABC, Generic[T]):
    """Abstract definition basic interface for a job class."""

    __type: JobType
    __timing: TimingJobUnion
    __handle: T
    __args: tuple[Any, ...]
    __kwargs: dict[str, Any]
    __max_attempts: int
    __tags: set[str]
    __delay: bool
    __start: Optional[dt.datetime]
    __stop: Optional[dt.datetime]
    __skip_missing: bool
    __alias: Optional[str]
    __tzinfo: Optional[dt.tzinfo]
    __logger: Logger

    __mark_delete: bool
    __attempts: int
    __failed_attempts: int
    __pending_timer: JobTimer
    __timers: list[JobTimer]

    def __init__(
        self,
        job_type: JobType,
        timing: TimingJobUnion,
        handle: T,
        *,
        args: Optional[tuple[Any, ...]] = None,
        kwargs: Optional[dict[str, Any]] = None,
        max_attempts: int = 0,
        tags: Optional[set[str]] = None,
        delay: bool = True,
        start: Optional[dt.datetime] = None,
        stop: Optional[dt.datetime] = None,
        skip_missing: bool = False,
        alias: Optional[str] = None,
        tzinfo: Optional[dt.tzinfo] = None,
    ):
        timing = standardize_timing_format(job_type, timing)

        sane_timing_types(job_type, timing)
        check_timing_tzinfo(job_type, timing, tzinfo)
        check_duplicate_effective_timings(job_type, timing, tzinfo)

        self.__start = set_start_check_stop_tzinfo(start, stop, tzinfo)

        self.__type = job_type
        self.__timing = timing  # pylint: disable=unused-private-member
        # NOTE: https://github.com/python/mypy/issues/708
        #       https://github.com/python/mypy/issues/2427
        self.__handle = handle
        self.__args = () if args is None else args
        self.__kwargs = {} if kwargs is None else kwargs.copy()
        self.__max_attempts = max_attempts
        self.__tags = set() if tags is None else tags.copy()
        self.__delay = delay
        self.__stop = stop
        self.__skip_missing = skip_missing
        self.__alias = alias
        self.__tzinfo = tzinfo

        # self.__mark_delete will be set to True if the new Timer would be in future
        # relativ to the self.__stop variable
        self.__mark_delete = False
        self.__attempts = 0
        self.__failed_attempts = 0

        # create JobTimers
        self.__timers = [JobTimer(job_type, tim, self.__start, skip_missing) for tim in timing]
        self.__pending_timer = get_pending_timer(self.__timers)

        if self.__stop is not None:
            if self.__pending_timer.datetime > self.__stop:
                self.__mark_delete = True

    def __lt__(self, other: BaseJob[T]) -> bool:
        return self.datetime < other.datetime

    def _calc_next_exec(self, ref_dt: dt.datetime) -> None:
        """
        Calculate the next estimated execution `datetime.datetime` of the `Job`.

        Parameters
        ----------
        ref_dt : datetime.datetime
            Reference time stamp to which the |BaseJob| calculates
            it's next execution.
        """
        if self.__skip_missing:
            for timer in self.__timers:
                if (timer.datetime - ref_dt).total_seconds() <= 0:
                    timer.calc_next_exec(ref_dt)
        else:
            self.__pending_timer.calc_next_exec(ref_dt)
        self.__pending_timer = get_pending_timer(self.__timers)
        if self.__stop is not None and self.__pending_timer.datetime > self.__stop:
            self.__mark_delete = True

    def _repr(self) -> tuple[str, ...]:
        return tuple(
            repr(elem)
            for elem in (
                self.__type,
                self.__timing,
                self.__handle,
                self.__args,
                self.__kwargs,
                self.__max_attempts,
                self.__delay,
                self.__start,
                self.__stop,
                self.__skip_missing,
                self.__alias,
                self.tzinfo,
            )
        )

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError()  # pragma: no cover

    def _str(
        self,
    ) -> tuple[str, ...]:
        """Return the objects relevant for readable string representation."""
        dt_timedelta = self.timedelta(dt.datetime.now(self.tzinfo))
        if self.alias is not None:
            f_args = ""
        elif hasattr(self.handle, "__code__"):
            f_args = "(..)" if self.handle.__code__.co_nlocals else "()"
        else:
            f_args = "(?)"
        return (
            self.type.name if self.max_attempts != 1 else "ONCE",
            self.handle.__qualname__ if self.alias is None else self.alias,
            f_args,
            str(self.datetime)[:19],
            str(self.datetime.tzname()),
            prettify_timedelta(dt_timedelta),
            str(self.attempts),
            str(float("inf") if self.max_attempts == 0 else self.max_attempts),
        )

    def __str__(self) -> str:
        return "{0}, {1}{2}, at={3}, tz={4}, in={5}, #{6}/{7}".format(*self._str())

    def timedelta(self, dt_stamp: Optional[dt.datetime] = None) -> dt.timedelta:
        """
        Get the `datetime.timedelta` until the next execution of this `Job`.

        Parameters
        ----------
        dt_stamp : Optional[datetime.datetime]
            Time to be compared with the planned execution time to determine the time difference.

        Returns
        -------
        timedelta
            `datetime.timedelta` to the next execution.
        """
        if dt_stamp is None:
            dt_stamp = dt.datetime.now(self.__tzinfo)
        if not self.__delay and self.__attempts == 0:
            return cast(dt.datetime, self.__start) - dt_stamp
        return self.__pending_timer.timedelta(dt_stamp)

    @property
    def datetime(self) -> dt.datetime:
        """
        Give the `datetime.datetime` object for the planed execution.

        Returns
        -------
        datetime.datetime
            Execution `datetime.datetime` stamp.
        """
        if not self.__delay and self.__attempts == 0:
            return cast(dt.datetime, self.__start)
        return self.__pending_timer.datetime

    @property
    def type(self) -> JobType:
        """
        Return the `JobType` of the `Job` instance.

        Returns
        -------
        JobType
            :class:`~scheduler.job.JobType` of the |BaseJob|.
        """
        return self.__type

    @property
    def handle(self) -> T:
        """
        Get the callback function.

        Returns
        -------
        handle
            Callback function.
        """
        return self.__handle

    @property
    def args(self) -> tuple[Any, ...]:
        r"""
        Get the positional arguments of the function handle within a `Job`.

        .. warning:: When running |BaseJob|\ s in parallel threads,
            be sure to implement possible side effects of parameter accessing in a
            thread safe manner.

        Returns
        -------
        tuple[Any]
            The payload arguments to pass to the function handle within a
            |BaseJob|.
        """
        return self.__args

    @property
    def kwargs(self) -> dict[str, Any]:
        r"""
        Get the keyword arguments of the function handle within a `Job`.

        .. warning:: When running |BaseJob|\ s in parallel threads,
            be sure to implement possible side effects of parameter accessing in a
            thread safe manner.

        Returns
        -------
        dict[str, Any]
            The payload arguments to pass to the function handle within a
            |BaseJob|.
        """
        return self.__kwargs

    @property
    def max_attempts(self) -> int:
        """
        Get the execution limit for a `Job`.

        Returns
        -------
        int
            Max execution attempts.
        """
        return self.__max_attempts

    @property
    def tags(self) -> set[str]:
        r"""
        Get the tags of a `Job`.

        Returns
        -------
        set[str]
            The tags of a |BaseJob|.
        """
        return self.__tags.copy()

    @property
    def delay(self) -> bool:
        """
        *Deprecated*: Return ``True`` if the first `Job` execution will wait for the next scheduled time.

        Returns
        -------
        bool
            If ``True`` wait with the execution for the next scheduled time. If ``False``
            the first execution will target the time of `Job.start`.
        """
        warnings.warn(
            (
                "Using the `delay` property is deprecated and will "
                "be removed in the next minor release."
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return self.__delay

    @property
    def start(self) -> Optional[dt.datetime]:
        """
        Get the timestamp at which the `JobTimer` starts.

        Returns
        -------
        Optional[datetime.datetime]
            The start datetime stamp.
        """
        return self.__start

    @property
    def stop(self) -> Optional[dt.datetime]:
        """
        Get the timestamp after which no more executions of the `Job` should be scheduled.

        Returns
        -------
        Optional[datetime.datetime]
            The stop datetime stamp.
        """
        return self.__stop

    @property
    def skip_missing(self) -> bool:
        """
        Return ``True`` if `Job` will only schedule it's newest planned execution.

        Returns
        -------
        bool
            If ``True`` a |BaseJob| will only schedule it's newest planned
            execution and drop older ones.
        """
        return self.__skip_missing

    @property
    def alias(self) -> Optional[str]:
        r"""
        Get the alias of the `Job`.

        Returns
        -------
        Optional[str]
            Alias of the |BaseJob|.
        """
        return self.__alias

    @property
    def tzinfo(self) -> Optional[dt.tzinfo]:
        r"""
        Get the timezone of the `Job`'s next execution.

        Returns
        -------
        Optional[datetime.tzinfo]
            Timezone of the |BaseJob|\ s next execution.
        """
        return self.datetime.tzinfo

    @property
    def _tzinfo(self) -> Optional[dt.tzinfo]:
        """
        Get the timezone of the `Scheduler` in which the `Job` is living.

        Returns
        -------
        Optional[datetime.tzinfo]
            Timezone of the |BaseJob|.
        """
        return self.__tzinfo

    @property
    def attempts(self) -> int:
        """
        Get the number of executions for a `Job`.

        Returns
        -------
        int
            Execution attempts.
        """
        return self.__attempts

    @property
    def failed_attempts(self) -> int:
        """
        Get the number of failed executions for a `Job`.

        Returns
        -------
        int
            Failed execution attempts.
        """
        return self.__failed_attempts

    @property
    def has_attempts_remaining(self) -> bool:
        """
        Check if a `Job` has remaining attempts.

        This function will return True if the |BaseJob| has open
        execution counts and the stop argument is not in the past relative to the
        next planed execution.

        Returns
        -------
        bool
            True if the |BaseJob| has execution attempts.
        """
        if self.__mark_delete:
            return False
        if self.__max_attempts == 0:
            return True
        return self.__attempts < self.__max_attempts


BaseJobType = TypeVar("BaseJobType", bound=BaseJob[Any])



import datetime as dt
from typing import Optional

import pytest

import scheduler.trigger as trigger
from scheduler import SchedulerError
from scheduler.base.definition import JobType
from scheduler.base.job_timer import JobTimer
from scheduler.base.job_util import sane_timing_types
from scheduler.base.timingtype import TimingJobTimerUnion, TimingJobUnion

from .helpers import CYCLIC_TYPE_ERROR_MSG, T_2021_5_26__3_55, utc


@pytest.mark.parametrize(
    "job_type, timing, start, target, next_target",
    (
        [
            JobType.WEEKLY,
            trigger.Thursday(),
            dt.datetime(year=2021, month=5, day=26, hour=11, minute=39, tzinfo=utc),
            dt.datetime(year=2021, month=5, day=27, tzinfo=utc),
            dt.datetime(year=2021, month=6, day=3, tzinfo=utc),
        ],
        [
            JobType.WEEKLY,
            trigger.Friday(dt.time(hour=1, minute=1, tzinfo=utc)),
            dt.datetime(year=2021, month=5, day=26, hour=11, minute=39, tzinfo=utc),
            dt.datetime(year=2021, month=5, day=28, hour=1, minute=1, tzinfo=utc),
            dt.datetime(year=2021, month=6, day=4, hour=1, minute=1, tzinfo=utc),
        ],
        [
            JobType.DAILY,
            dt.time(hour=12, minute=1, tzinfo=utc),
            dt.datetime(year=2021, month=5, day=26, hour=11, minute=39, tzinfo=utc),
            dt.datetime(year=2021, month=5, day=26, hour=12, minute=1, tzinfo=utc),
            dt.datetime(year=2021, month=5, day=27, hour=12, minute=1, tzinfo=utc),
        ],
        [
            JobType.HOURLY,
            dt.time(minute=1, tzinfo=utc),
            dt.datetime(year=2021, month=5, day=26, hour=11, minute=39, tzinfo=utc),
            dt.datetime(year=2021, month=5, day=26, hour=12, minute=1, tzinfo=utc),
            dt.datetime(year=2021, month=5, day=26, hour=13, minute=1, tzinfo=utc),
        ],
        [
            JobType.MINUTELY,
            dt.time(second=1, tzinfo=utc),
            dt.datetime(year=2021, month=5, day=26, hour=11, minute=39, tzinfo=utc),
            dt.datetime(year=2021, month=5, day=26, hour=11, minute=39, second=1, tzinfo=utc),
            dt.datetime(year=2021, month=5, day=26, hour=11, minute=40, second=1, tzinfo=utc),
        ],
        [
            JobType.CYCLIC,
            dt.timedelta(hours=3, seconds=27),
            dt.datetime(year=2021, month=5, day=26, hour=11, minute=39),
            dt.datetime(year=2021, month=5, day=26, hour=11, minute=39)
            + dt.timedelta(hours=3, seconds=27),
            dt.datetime(year=2021, month=5, day=26, hour=11, minute=39)
            + 2 * dt.timedelta(hours=3, seconds=27),
        ],
    ),
)
def test_JobTimer_calc_next_exec(
    job_type: JobType,
    timing: TimingJobTimerUnion,
    start: dt.datetime,
    target: dt.datetime,
    next_target: dt.datetime,
) -> None:
    timer = JobTimer(job_type, timing, start)

    assert timer.datetime == target
    assert timer.timedelta(start) == target - start

    timer.calc_next_exec()
    assert timer.datetime == next_target


@pytest.mark.parametrize(
    "delta_m, offset_m, skip, res_delta_m",
    (
        [20, 21, True, 41],
        [1, 21, True, 22],
        [1, 1, True, 2],
        [20, 20, True, 40],
        [20, 1, True, 21],
        [20, 21, False, 40],
        [1, 21, False, 2],
        [1, 1, False, 2],
        [20, 20, False, 40],
        [20, 1, False, 40],
    ),
)
def test_skip(delta_m: int, offset_m: int, skip: bool, res_delta_m: int) -> None:
    delta = dt.timedelta(minutes=delta_m)
    offset = dt.timedelta(minutes=offset_m)
    res_delta = dt.timedelta(minutes=res_delta_m)

    jet = JobTimer(
        JobType.CYCLIC,
        timing=delta,
        start=T_2021_5_26__3_55,
        skip_missing=skip,
    )
    assert jet.datetime == T_2021_5_26__3_55 + delta

    jet.calc_next_exec(T_2021_5_26__3_55 + offset)
    assert jet.datetime == T_2021_5_26__3_55 + res_delta


@pytest.mark.parametrize(
    "job_type, timing, err",
    (
        [JobType.CYCLIC, [dt.timedelta()], None],
        [JobType.CYCLIC, [dt.timedelta(), dt.timedelta()], CYCLIC_TYPE_ERROR_MSG],
        [JobType.WEEKLY, [trigger.Monday()], None],
        [JobType.DAILY, [dt.time()], None],
        [JobType.DAILY, [dt.time(), dt.time()], None],
        [JobType.HOURLY, [dt.time()], None],
        [JobType.HOURLY, [dt.time(), dt.time()], None],
        [JobType.MINUTELY, [dt.time()], None],
        [JobType.MINUTELY, [dt.time(), dt.time()], None],
        # [JobType.CYCLIC, (dt.timedelta(), dt.timedelta()), CYCLIC_TYPE_ERROR_MSG],
        # [JobType.WEEKLY, dt.time(), WEEKLY_TYPE_ERROR_MSG],
        # [JobType.WEEKLY, [trigger.Monday(), dt.time()], WEEKLY_TYPE_ERROR_MSG],
        # [JobType.DAILY, (dt.time(), dt.time()), DAILY_TYPE_ERROR_MSG],
        # [JobType.HOURLY, (dt.time(), dt.time()), HOURLY_TYPE_ERROR_MSG],
        # [JobType.MINUTELY, (dt.time(), dt.time()), MINUTELY_TYPE_ERROR_MSG],
    ),
)
def test_sane_timing_types(job_type: JobType, timing: TimingJobUnion, err: Optional[str]) -> None:
    if err:
        with pytest.raises(SchedulerError, match=err):
            sane_timing_types(job_type, timing)
    else:
        sane_timing_types(job_type, timing)



from scheduler.trigger import (
    Friday,
    Monday,
    Saturday,
    Sunday,
    Thursday,
    Tuesday,
    Wednesday,
    weekday,
)

from .helpers import samples, samples_utc


def test_trigger_misc() -> None:
    for sample in samples + samples_utc:
        for day, wkday in zip(
            range(7), (Monday, Tuesday, Wednesday, Thursday, Friday, Saturday, Sunday)
        ):

            res = weekday(value=day, time=sample.time())
            assert isinstance(res, wkday)
            assert res.value == day
            assert res.time == sample.time()



"""
Implementation of the essential timer for a `BaseJob`.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

from __future__ import annotations

import datetime as dt
import threading
from typing import Optional, cast

from scheduler.base.definition import JobType
from scheduler.base.timingtype import TimingJobTimerUnion
from scheduler.trigger.core import Weekday
from scheduler.util import JOB_NEXT_DAYLIKE_MAPPING, next_weekday_time_occurrence


class JobTimer:
    """
    The class provides the internal `datetime.datetime` calculations for a |BaseJob|.

    Parameters
    ----------
    job_type : JobType
        Indicator which defines which calculations has to be used.
    timing : TimingJobTimerUnion
        Desired execution time(s).
    start : datetime.datetime
        Timestamp reference from which future executions will be calculated.
    skip_missing : bool
        If ``True`` a |BaseJob| will only schedule it's newest planned
        execution and drop older ones.
    """

    def __init__(
        self,
        job_type: JobType,
        timing: TimingJobTimerUnion,
        start: dt.datetime,
        skip_missing: bool = False,
    ):
        self.__lock = threading.RLock()
        self.__job_type = job_type
        self.__timing = timing
        self.__next_exec = start
        self.__skip = skip_missing
        self.calc_next_exec()

    def calc_next_exec(self, ref: Optional[dt.datetime] = None) -> None:
        """
        Generate the next execution `datetime.datetime` stamp.

        Parameters
        ----------
        ref : Optional[datetime.datetime]
            Datetime reference for scheduling the next execution datetime.
        """
        with self.__lock:
            if self.__job_type == JobType.CYCLIC:
                if self.__skip and ref is not None:
                    self.__next_exec = ref
                self.__next_exec = self.__next_exec + cast(dt.timedelta, self.__timing)
                return

            if self.__job_type == JobType.WEEKLY:
                self.__timing = cast(Weekday, self.__timing)
                if self.__timing.time.tzinfo:
                    self.__next_exec = self.__next_exec.astimezone(self.__timing.time.tzinfo)
                self.__next_exec = next_weekday_time_occurrence(
                    self.__next_exec, self.__timing, self.__timing.time
                )

            else:  # self.__job_type in JOB_NEXT_DAYLIKE_MAPPING:
                self.__timing = cast(dt.time, self.__timing)
                if self.__next_exec.tzinfo:
                    self.__next_exec = self.__next_exec.astimezone(self.__timing.tzinfo)
                self.__next_exec = JOB_NEXT_DAYLIKE_MAPPING[self.__job_type](
                    self.__next_exec, self.__timing
                )

            if self.__skip and ref is not None and self.__next_exec < ref:
                self.__next_exec = ref
                self.calc_next_exec()

    @property
    def datetime(self) -> dt.datetime:
        """
        Get the `datetime.datetime` object for the planed execution.

        Returns
        -------
        datetime.datetime
            Execution `datetime.datetime` stamp.
        """
        with self.__lock:
            return self.__next_exec

    def timedelta(self, dt_stamp: dt.datetime) -> dt.timedelta:
        """
        Get the `datetime.timedelta` until the execution of this `Job`.

        Parameters
        ----------
        dt_stamp : datetime.datetime
            Time to be compared with the planned execution time
            to determine the time difference.

        Returns
        -------
        datetime.timedelta
            `datetime.timedelta` to the execution.
        """
        with self.__lock:
            return self.__next_exec - dt_stamp



import datetime as dt
from typing import Callable

import pytest

import scheduler
from scheduler.prioritization import (
    constant_weight_prioritization,
    linear_priority_function,
    random_priority_function,
)
from scheduler.threading.job import Job


@pytest.mark.parametrize(
    "timedelta, executions",
    [
        [dt.timedelta(seconds=0), 1],
        [dt.timedelta(seconds=100), 0],
    ],
)
@pytest.mark.parametrize(
    "priority_function",
    [
        constant_weight_prioritization,
        linear_priority_function,
    ],
)
def test_deprecated_prioritization(
    timedelta: dt.timedelta,
    executions: int,
    priority_function: Callable[
        [float, Job, int, int],
        float,
    ],
) -> None:
    schedule = scheduler.Scheduler(max_exec=3, priority_function=priority_function)
    schedule.once(
        dt.datetime.now() + timedelta,
        print,
    )
    assert schedule.exec_jobs() == executions


@pytest.mark.parametrize(
    "timedelta, weight, executions",
    [
        [dt.timedelta(seconds=0), 1, 1],
        [dt.timedelta(seconds=0), 0, 0],
        [dt.timedelta(seconds=100), 1, 1],
    ],
)
def test_deprecated_rnd_prioritization(
    timedelta: dt.timedelta, weight: int, executions: int
) -> None:
    schedule = scheduler.Scheduler(max_exec=3, priority_function=random_priority_function)
    schedule.once(dt.datetime.now() + timedelta, print, weight=weight)
    assert schedule.exec_jobs() == executions



"""
Implementation of a `threading` compatible in-process scheduler.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

import datetime as dt
import queue
import threading
from collections.abc import Iterable
from logging import Logger
from typing import Any, Callable, Optional

import typeguard as tg

from scheduler.base.definition import JOB_TYPE_MAPPING, JobType
from scheduler.base.scheduler import BaseScheduler, deprecated, select_jobs_by_tag
from scheduler.base.scheduler_util import check_tzname, create_job_instance, str_cutoff
from scheduler.base.timingtype import (
    TimingCyclic,
    TimingDailyUnion,
    TimingOnceUnion,
    TimingWeeklyUnion,
)
from scheduler.error import SchedulerError
from scheduler.message import (
    CYCLIC_TYPE_ERROR_MSG,
    DAILY_TYPE_ERROR_MSG,
    HOURLY_TYPE_ERROR_MSG,
    MINUTELY_TYPE_ERROR_MSG,
    ONCE_TYPE_ERROR_MSG,
    TZ_ERROR_MSG,
    WEEKLY_TYPE_ERROR_MSG,
)
from scheduler.prioritization import linear_priority_function
from scheduler.threading.job import Job


def _exec_job_worker(que: queue.Queue[Job], logger: Logger) -> None:
    running = True
    while running:
        try:
            job = que.get(block=False)
        except queue.Empty:
            running = False
        else:
            job._exec(logger=logger)  # pylint: disable=protected-access
            que.task_done()


class Scheduler(BaseScheduler[Job, Callable[..., None]]):
    r"""
    Implementation of a scheduler for callback functions.

    This implementation enables the planning of |Job|\ s depending on time
    cycles, fixed times, weekdays, dates, offsets, execution counts and weights.

    Notes
    -----
    Due to the support of `datetime` objects, `scheduler` is able to work
    with timezones.

    Parameters
    ----------
    tzinfo : datetime.tzinfo
        Set the timezone of the |Scheduler|.
    max_exec : int
        Limits the number of overdue |Job|\ s that can be executed
        by calling function `Scheduler.exec_jobs()`.
    priority_function : Callable[[float, Job, int, int], float]
        A function handle to compute the priority of a |Job| depending
        on the time it is overdue and its respective weight. Defaults to a linear
        priority function.
    jobs : set[Job]
        A collection of job instances.
    n_threads : int
        The number of worker threads. 0 for unlimited, default 1.
    logger : Optional[logging.Logger]
        A custom Logger instance.
    """

    def __init__(
        self,
        *,
        max_exec: int = 0,
        tzinfo: Optional[dt.tzinfo] = None,
        priority_function: Callable[
            [float, Job, int, int],
            float,
        ] = linear_priority_function,
        jobs: Optional[Iterable[Job]] = None,
        n_threads: int = 1,
        logger: Optional[Logger] = None,
    ):
        super().__init__(logger=logger)
        self.__max_exec = max_exec
        self.__tzinfo = tzinfo
        self.__priority_function = priority_function
        self.__jobs_lock = threading.RLock()
        if not jobs:
            self.__jobs = set()
        elif isinstance(jobs, set):
            self.__jobs = jobs
        else:
            self.__jobs = set(jobs)

        for job in self.__jobs:
            if job._tzinfo != self.__tzinfo:
                raise SchedulerError(TZ_ERROR_MSG)

        self.__n_threads = n_threads
        self.__tz_str = check_tzname(tzinfo=tzinfo)

    def __repr__(self) -> str:
        with self.__jobs_lock:
            return "scheduler.Scheduler({0}, jobs={{{1}}})".format(
                ", ".join(
                    (
                        repr(elem)
                        for elem in (
                            self.__max_exec,
                            self.__tzinfo,
                            self.__priority_function,
                        )
                    )
                ),
                ", ".join([repr(job) for job in sorted(self.jobs)]),
            )

    def __str__(self) -> str:
        with self.__jobs_lock:
            # Scheduler meta heading
            scheduler_headings = "{0}, {1}, {2}, {3}\n\n".format(*self.__headings())

            # Job table (we join two of the Job._repr() fields into one)
            # columns
            c_align = ("<", "<", "<", "<", ">", ">", ">")
            c_width = (8, 16, 19, 12, 9, 13, 6)
            c_name = (
                "type",
                "function / alias",
                "due at",
                "tzinfo",
                "due in",
                "attempts",
                "weight",
            )
            form = [
                f"{{{idx}:{align}{width}}}"
                for idx, (align, width) in enumerate(zip(c_align, c_width))
            ]
            if self.__tz_str is None:
                form = form[:3] + form[4:]

            fstring = " ".join(form) + "\n"
            job_table = fstring.format(*c_name) + fstring.format(
                *("-" * width for width in c_width)
            )
            for job in sorted(self.jobs):
                row = job._str()
                entries = (
                    row[0],
                    str_cutoff(row[1] + row[2], c_width[1], False),
                    row[3],
                    str_cutoff(row[4] or "", c_width[3], False),
                    str_cutoff(row[5], c_width[4], True),
                    str_cutoff(f"{row[6]}/{row[7]}", c_width[5], True),
                    str_cutoff(f"{job.weight}", c_width[6], True),
                )
                job_table += fstring.format(*entries)

            return scheduler_headings + job_table

    def __headings(self) -> list[str]:
        with self.__jobs_lock:
            headings = [
                f"max_exec={self.__max_exec if self.__max_exec else float('inf')}",
                f"tzinfo={self.__tz_str}",
                f"priority_function={self.__priority_function.__name__}",
                f"#jobs={len(self.__jobs)}",
            ]
            return headings

    def __schedule(
        self,
        **kwargs,
    ) -> Job:
        """Encapsulate the `Job` and add the `Scheduler`'s timezone."""
        job: Job = create_job_instance(Job, tzinfo=self.__tzinfo, **kwargs)
        if job.has_attempts_remaining:
            with self.__jobs_lock:
                self.__jobs.add(job)
        return job

    def __exec_jobs(self, jobs: list[Job], ref_dt: dt.datetime) -> int:
        n_jobs = len(jobs)

        que: queue.Queue[Job] = queue.Queue()
        for job in jobs:
            que.put(job)

        workers = []
        for _ in range(self.__n_threads or n_jobs):
            worker = threading.Thread(target=_exec_job_worker, args=(que, self._logger))
            worker.daemon = True
            worker.start()
            workers.append(worker)

        que.join()
        for worker in workers:
            worker.join()

        for job in jobs:
            job._calc_next_exec(ref_dt)  # pylint: disable=protected-access
            if not job.has_attempts_remaining:
                self.delete_job(job)

        return n_jobs

    def exec_jobs(self, force_exec_all: bool = False) -> int:
        r"""
        Execute scheduled `Job`\ s.

        By default executes the |Job|\ s that are overdue.

        |Job|\ s are executed in order of their priority
        :ref:`examples.weights`. If the |Scheduler| instance
        has a limit on the job execution counts per call of
        :func:`~scheduler.core.Scheduler.exec_jobs`, via the `max_exec` argument,
        |Job|\ s of lower priority might not get executed when
        competing |Job|\ s are overdue.

        Parameters
        ----------
        force_exec_all : bool
            Ignore the both - the status of the |Job| timers
            as well as the execution limit of the |Scheduler|

        Returns
        -------
        int
            Number of executed |Job|\ s.
        """
        ref_dt = dt.datetime.now(tz=self.__tzinfo)

        if force_exec_all:
            return self.__exec_jobs(list(self.__jobs), ref_dt)
        #  collect the current priority for all jobs

        job_priority: dict[Job, float] = {}
        n_jobs = len(self.__jobs)
        with self.__jobs_lock:
            for job in self.__jobs:
                delta_seconds = job.timedelta(ref_dt).total_seconds()
                job_priority[job] = self.__priority_function(
                    -delta_seconds,
                    job,
                    self.__max_exec,
                    n_jobs,
                )
        # sort the jobs by priority
        sorted_jobs = sorted(job_priority, key=job_priority.get, reverse=True)  # type: ignore
        # filter jobs by max_exec and priority greater zero
        filtered_jobs = [
            job
            for idx, job in enumerate(sorted_jobs)
            if (self.__max_exec == 0 or idx < self.__max_exec) and job_priority[job] > 0
        ]
        return self.__exec_jobs(filtered_jobs, ref_dt)

    def delete_job(self, job: Job) -> None:
        """
        Delete a `Job` from the `Scheduler`.

        Parameters
        ----------
        job : Job
            |Job| instance to delete.

        Raises
        ------
        SchedulerError
            Raises if the |Job| of the argument is not scheduled.
        """
        try:
            with self.__jobs_lock:
                self.__jobs.remove(job)
        except KeyError:
            raise SchedulerError("An unscheduled Job can not be deleted!") from None

    def delete_jobs(
        self,
        tags: Optional[set[str]] = None,
        any_tag: bool = False,
    ) -> int:
        r"""
        Delete a set of |Job|\ s from the |Scheduler| by tags.

        If no tags or an empty set of tags are given defaults to the deletion
        of all |Job|\ s.

        Parameters
        ----------
        tags : Optional[set[str]]
            Set of tags to identify target |Job|\ s.
        any_tag : bool
            False: To delete a |Job| all tags have to match.
            True: To deleta a |Job| at least one tag has to match.
        """
        with self.__jobs_lock:
            if tags is None or tags == set():
                n_jobs = len(self.__jobs)
                self.__jobs = set()
                return n_jobs

            to_delete = select_jobs_by_tag(self.__jobs, tags, any_tag)

            self.__jobs = self.__jobs - to_delete
            return len(to_delete)

    def get_jobs(
        self,
        tags: Optional[set[str]] = None,
        any_tag: bool = False,
    ) -> set[Job]:
        r"""
        Get a set of |Job|\ s from the |Scheduler| by tags.

        If no tags or an empty set of tags are given defaults to returning
        all |Job|\ s.

        Parameters
        ----------
        tags : set[str]
            Tags to filter scheduled |Job|\ s.
            If no tags are given all |Job|\ s are returned.
        any_tag : bool
            False: To match a |Job| all tags have to match.
            True: To match a |Job| at least one tag has to match.

        Returns
        -------
        set[Job]
            Currently scheduled |Job|\ s.
        """
        with self.__jobs_lock:
            if tags is None or tags == set():
                return self.__jobs.copy()
            return select_jobs_by_tag(self.__jobs, tags, any_tag)

    @deprecated(["delay"])
    def cyclic(self, timing: TimingCyclic, handle: Callable[..., None], **kwargs) -> Job:
        r"""
        Schedule a cyclic `Job`.

        Use a `datetime.timedelta` object or a `list` of `datetime.timedelta` objects
        to schedule a cyclic |Job|.

        Parameters
        ----------
        timing : TimingTypeCyclic
            Desired execution time.
        handle : Callable[..., None]
            Handle to a callback function.

        Returns
        -------
        Job
            Instance of a scheduled |Job|.

        Other Parameters
        ----------------
        **kwargs
            |Job| properties, optional

            `kwargs` are used to specify |Job| properties.

            Here is a list of available |Job| properties:

            .. include:: ../_assets/kwargs.rst
        """
        try:
            tg.check_type(timing, TimingCyclic)
        except tg.TypeCheckError as err:
            raise SchedulerError(CYCLIC_TYPE_ERROR_MSG) from err
        return self.__schedule(job_type=JobType.CYCLIC, timing=timing, handle=handle, **kwargs)

    @deprecated(["delay"])
    def minutely(self, timing: TimingDailyUnion, handle: Callable[..., None], **kwargs) -> Job:
        r"""
        Schedule a minutely `Job`.

        Use a `datetime.time` object or a `list` of `datetime.time` objects
        to schedule a |Job| every minute.

        Notes
        -----
        If given a `datetime.time` object with a non zero hour or minute property, these
        information will be ignored.

        Parameters
        ----------
        timing : TimingDailyUnion
            Desired execution time(s).
        handle : Callable[..., None]
            Handle to a callback function.

        Returns
        -------
        Job
            Instance of a scheduled |Job|.

        Other Parameters
        ----------------
        **kwargs
            |Job| properties, optional

            `kwargs` are used to specify |Job| properties.

            Here is a list of available |Job| properties:

            .. include:: ../_assets/kwargs.rst
        """
        try:
            tg.check_type(timing, TimingDailyUnion)
        except tg.TypeCheckError as err:
            raise SchedulerError(MINUTELY_TYPE_ERROR_MSG) from err
        return self.__schedule(job_type=JobType.MINUTELY, timing=timing, handle=handle, **kwargs)

    @deprecated(["delay"])
    def hourly(self, timing: TimingDailyUnion, handle: Callable[..., None], **kwargs) -> Job:
        r"""
        Schedule an hourly `Job`.

        Use a `datetime.time` object or a `list` of `datetime.time` objects
        to schedule a |Job| every hour.

        Notes
        -----
        If given a `datetime.time` object with a non zero hour property, this information
        will be ignored.

        Parameters
        ----------
        timing : TimingDailyUnion
            Desired execution time(s).
        handle : Callable[..., None]
            Handle to a callback function.

        Returns
        -------
        Job
            Instance of a scheduled |Job|.

        Other Parameters
        ----------------
        **kwargs
            |Job| properties, optional

            `kwargs` are used to specify |Job| properties.

            Here is a list of available |Job| properties:

            .. include:: ../_assets/kwargs.rst
        """
        try:
            tg.check_type(timing, TimingDailyUnion)
        except tg.TypeCheckError as err:
            raise SchedulerError(HOURLY_TYPE_ERROR_MSG) from err
        return self.__schedule(job_type=JobType.HOURLY, timing=timing, handle=handle, **kwargs)

    @deprecated(["delay"])
    def daily(self, timing: TimingDailyUnion, handle: Callable[..., None], **kwargs) -> Job:
        r"""
        Schedule a daily `Job`.

        Use a `datetime.time` object or a `list` of `datetime.time` objects
        to schedule a |Job| every day.

        Parameters
        ----------
        timing : TimingDailyUnion
            Desired execution time(s).
        handle : Callable[..., None]
            Handle to a callback function.

        Returns
        -------
        Job
            Instance of a scheduled |Job|.

        Other Parameters
        ----------------
        **kwargs
            |Job| properties, optional

            `kwargs` are used to specify |Job| properties.

            Here is a list of available |Job| properties:

            .. include:: ../_assets/kwargs.rst
        """
        try:
            tg.check_type(timing, TimingDailyUnion)
        except tg.TypeCheckError as err:
            raise SchedulerError(DAILY_TYPE_ERROR_MSG) from err
        return self.__schedule(job_type=JobType.DAILY, timing=timing, handle=handle, **kwargs)

    @deprecated(["delay"])
    def weekly(self, timing: TimingWeeklyUnion, handle: Callable[..., None], **kwargs) -> Job:
        r"""
        Schedule a weekly `Job`.

        Use a `tuple` of a `Weekday` and a `datetime.time` object to define a weekly
        recurring |Job|. Combine multiple desired `tuples` in
        a `list`. If the planed execution time is `00:00` the `datetime.time` object
        can be ignored, just pass a `Weekday` without a `tuple`.

        Parameters
        ----------
        timing : TimingWeeklyUnion
            Desired execution time(s).
        handle : Callable[..., None]
            Handle to a callback function.

        Returns
        -------
        Job
            Instance of a scheduled |Job|.

        Other Parameters
        ----------------
        **kwargs
            |Job| properties, optional

            `kwargs` are used to specify |Job| properties.

            Here is a list of available |Job| properties:

            .. include:: ../_assets/kwargs.rst
        """
        try:
            tg.check_type(timing, TimingWeeklyUnion)
        except tg.TypeCheckError as err:
            raise SchedulerError(WEEKLY_TYPE_ERROR_MSG) from err
        return self.__schedule(job_type=JobType.WEEKLY, timing=timing, handle=handle, **kwargs)

    def once(  # pylint: disable=arguments-differ
        self,
        timing: TimingOnceUnion,
        handle: Callable[..., None],
        *,
        args: Optional[tuple[Any]] = None,
        kwargs: Optional[dict[str, Any]] = None,
        tags: Optional[Iterable[str]] = None,
        alias: Optional[str] = None,
        weight: float = 1,
    ) -> Job:
        r"""
        Schedule a oneshot `Job`.

        Parameters
        ----------
        timing : TimingOnceUnion
            Desired execution time.
        handle : Callable[..., None]
            Handle to a callback function.
        args : tuple[Any]
            Positional argument payload for the function handle within a |Job|.
        kwargs : Optional[dict[str, Any]]
            Keyword arguments payload for the function handle within a |Job|.
        tags : Optional[Iterable[str]]
            The tags of the |Job|.
        alias : Optional[str]
            Overwrites the function handle name in the string representation.
        weight : float
            Relative weight against other |Job|\ s.

        Returns
        -------
        Job
            Instance of a scheduled |Job|.
        """
        try:
            tg.check_type(timing, TimingOnceUnion)
        except tg.TypeCheckError as err:
            raise SchedulerError(ONCE_TYPE_ERROR_MSG) from err
        if isinstance(timing, dt.datetime):
            return self.__schedule(
                job_type=JobType.CYCLIC,
                timing=dt.timedelta(),
                handle=handle,
                args=args,
                kwargs=kwargs,
                max_attempts=1,
                tags=set(tags) if tags else set(),
                alias=alias,
                weight=weight,
                delay=False,
                start=timing,
            )
        return self.__schedule(
            job_type=JOB_TYPE_MAPPING[type(timing)],
            timing=timing,
            handle=handle,
            args=args,
            kwargs=kwargs,
            max_attempts=1,
            tags=tags,
            alias=alias,
            weight=weight,
        )

    @property
    def jobs(self) -> set[Job]:
        r"""
        Get the set of all `Job`\ s.

        Returns
        -------
        set[Job]
            Currently scheduled |Job|\ s.
        """
        return self.__jobs.copy()



import datetime as dt
from typing import Optional, Union

import pytest

import scheduler.trigger as trigger
from scheduler.base.scheduler_util import str_cutoff
from scheduler.error import SchedulerError
from scheduler.trigger.core import Weekday, _Weekday
from scheduler.util import (
    days_to_weekday,
    next_daily_occurrence,
    next_hourly_occurrence,
    next_minutely_occurrence,
    next_weekday_time_occurrence,
)

err_msg = r"Weekday enumeration interval: \[0,6\] <=> \[Monday, Sunday\]"


@pytest.mark.parametrize(
    "wkdy_src, wkdy_dest, days, err_msg",
    (
        [trigger.Monday(), trigger.Thursday(), 3, None],
        [trigger.Wednesday(), trigger.Sunday(), 4, None],
        [trigger.Friday(), trigger.Friday(), 7, None],
        [trigger.Saturday(), trigger.Thursday(), 5, None],
        [trigger.Sunday(), trigger.Saturday(), 6, None],
        [3, 8, 0, err_msg],
        [4, -1, 0, err_msg],
        [8, 4, 0, err_msg],
        [-1, 2, 0, err_msg],
    ),
)
def test_days_to_weekday(
    wkdy_src: Union[int, _Weekday], wkdy_dest: int, days: int, err_msg: Optional[str]
) -> None:
    if err_msg:
        assert isinstance(wkdy_src, int)
        with pytest.raises(SchedulerError, match=err_msg):
            days_to_weekday(wkdy_src, wkdy_dest)
    else:
        assert isinstance(wkdy_src, Weekday)
        assert isinstance(wkdy_dest, Weekday)

        assert days_to_weekday(wkdy_src.value, wkdy_dest.value) == days


@pytest.mark.parametrize(
    "now, wkdy, timestamp, target",
    (
        [
            dt.datetime(year=2021, month=5, day=26, hour=11, minute=39),
            trigger.Friday(),
            dt.time(hour=0, minute=0),
            dt.datetime(year=2021, month=5, day=28),
        ],
        [
            dt.datetime(year=2021, month=5, day=26, hour=11, minute=39),
            trigger.Wednesday(),
            dt.time(hour=12, minute=3, second=1),
            dt.datetime(year=2021, month=5, day=26, hour=12, minute=3, second=1),
        ],
        [
            dt.datetime(year=2021, month=5, day=26, hour=11, minute=39, tzinfo=dt.timezone.utc),
            trigger.Thursday(),
            dt.time(hour=12, minute=3, second=1, tzinfo=dt.timezone.utc),
            dt.datetime(
                year=2021,
                month=5,
                day=27,
                hour=12,
                minute=3,
                second=1,
                tzinfo=dt.timezone.utc,
            ),
        ],
        [
            dt.datetime(year=2021, month=6, day=16, hour=1, minute=53, second=45),
            trigger.Wednesday(),
            dt.time(hour=2),
            dt.datetime(year=2021, month=6, day=16, hour=2),
        ],
    ),
)
def test_next_weekday_time_occurrence(
    now: dt.datetime, wkdy: _Weekday, timestamp: dt.time, target: dt.datetime
) -> None:
    assert next_weekday_time_occurrence(now, wkdy, timestamp) == target


@pytest.mark.parametrize(
    "now, target_time, target_datetime",
    (
        [
            dt.datetime(year=2021, month=6, day=16, hour=1, minute=53, second=45),
            dt.time(hour=2),
            dt.datetime(year=2021, month=6, day=16, hour=2),
        ],
        [
            dt.datetime(year=2021, month=6, day=16, hour=1, minute=53, second=45),
            dt.time(hour=13, second=45),
            dt.datetime(year=2021, month=6, day=16, hour=13, second=45),
        ],
        [
            dt.datetime(year=2021, month=6, day=16, hour=1, minute=53, second=45),
            dt.time(second=45),
            dt.datetime(year=2021, month=6, day=17, second=45),
        ],
    ),
)
def test_next_daily_occurence(
    now: dt.datetime, target_time: dt.time, target_datetime: dt.datetime
) -> None:
    assert next_daily_occurrence(now, target_time) == target_datetime


@pytest.mark.parametrize(
    "now, target_time, target_datetime",
    (
        [
            dt.datetime(year=2021, month=6, day=16, hour=1, minute=53, second=45),
            dt.time(minute=7, second=3),
            dt.datetime(year=2021, month=6, day=16, hour=2, minute=7, second=3),
        ],
        [
            dt.datetime(year=2021, month=6, day=16, hour=23, minute=53, second=45),
            dt.time(minute=7, second=3),
            dt.datetime(year=2021, month=6, day=17, hour=0, minute=7, second=3),
        ],
        [
            dt.datetime(year=2021, month=6, day=16, hour=1, minute=53, second=45),
            dt.time(hour=4),
            dt.datetime(year=2021, month=6, day=16, hour=2),
        ],
    ),
)
def test_next_hourly_occurence(
    now: dt.datetime, target_time: dt.time, target_datetime: dt.datetime
) -> None:
    assert next_hourly_occurrence(now, target_time) == target_datetime


@pytest.mark.parametrize(
    "now, target_time, target_datetime",
    (
        [
            dt.datetime(year=2021, month=6, day=16, hour=1, minute=53, second=45),
            dt.time(second=3),
            dt.datetime(year=2021, month=6, day=16, hour=1, minute=54, second=3),
        ],
        [
            dt.datetime(year=2021, month=6, day=16, hour=23, minute=59, second=45),
            dt.time(second=44),
            dt.datetime(year=2021, month=6, day=17, hour=0, minute=0, second=44),
        ],
        [
            dt.datetime(year=2021, month=6, day=16, hour=1, minute=53, second=45),
            dt.time(hour=4, minute=8, second=25),
            dt.datetime(year=2021, month=6, day=16, hour=1, minute=54, second=25),
        ],
    ),
)
def test_next_minutely_occurence(
    now: dt.datetime, target_time: dt.time, target_datetime: dt.datetime
) -> None:
    assert next_minutely_occurrence(now, target_time) == target_datetime


@pytest.mark.parametrize(
    "string, max_length, cut_tail, result, err",
    [
        ("abcdefg", 10, False, "abcdefg", None),
        ("abcdefg", 4, False, "#efg", None),
        ("abcdefg", 2, True, "a#", None),
        ("abcdefg", 1, True, "#", None),
        ("abcdefg", 0, True, "", "max_length < 1 not allowed"),
    ],
)
def test_str_cutoff(
    string: str, max_length: int, cut_tail: bool, result: str, err: Optional[str]
) -> None:
    if err:
        with pytest.raises(ValueError, match=err):
            str_cutoff(string, max_length, cut_tail)
    else:
        assert str_cutoff(string, max_length, cut_tail) == result



"""
Collection of datetime and trigger related utility functions.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""
from __future__ import annotations

import datetime as dt
from typing import Optional

from scheduler.base.definition import JobType
from scheduler.error import SchedulerError
from scheduler.trigger.core import Weekday


def days_to_weekday(wkdy_src: int, wkdy_dest: int) -> int:
    """
    Calculate the days to a specific destination weekday.

    Notes
    -----
    Weekday enumeration based on
    the `datetime` standard library.

    Parameters
    ----------
    wkdy_src : int
        Source :class:`~scheduler.util.Weekday` integer representation.
    wkdy_dest : int
        Destination :class:`~scheduler.util.Weekday` integer representation.

    Returns
    -------
    int
        Days to the destination :class:`~scheduler.util.Weekday`.
    """
    if not (0 <= wkdy_src <= 6 and 0 <= wkdy_dest <= 6):
        raise SchedulerError("Weekday enumeration interval: [0,6] <=> [Monday, Sunday]")

    return (wkdy_dest - wkdy_src - 1) % 7 + 1


def next_daily_occurrence(now: dt.datetime, target_time: dt.time) -> dt.datetime:
    """
    Estimate the next daily occurrence of a given time.

    .. warning:: Both arguments are expected to have the same tzinfo, no internal checks.

    Parameters
    ----------
    now : datetime.datetime
        `datetime.datetime` object of today
    target_time : datetime.time
        Desired `datetime.time`.

    Returns
    -------
    datetime.datetime
        Next `datetime.datetime` object with the desired time.
    """
    target = now.replace(
        hour=target_time.hour,
        minute=target_time.minute,
        second=target_time.second,
        microsecond=target_time.microsecond,
    )
    if (target - now).total_seconds() <= 0:
        target = target + dt.timedelta(days=1)
    return target


def next_hourly_occurrence(now: dt.datetime, target_time: dt.time) -> dt.datetime:
    """
    Estimate the next hourly occurrence of a given time.

    .. warning:: Both arguments are expected to have the same tzinfo, no internal checks.

    Parameters
    ----------
    now : datetime.datetime
        `datetime.datetime` object of today
    target_time : datetime.time
        Desired `datetime.time`.

    Returns
    -------
    datetime.datetime
        Next `datetime.datetime` object with the desired time.
    """
    target = now.replace(
        minute=target_time.minute,
        second=target_time.second,
        microsecond=target_time.microsecond,
    )
    if (target - now).total_seconds() <= 0:
        target = target + dt.timedelta(hours=1)
    return target


def next_minutely_occurrence(now: dt.datetime, target_time: dt.time) -> dt.datetime:
    """
    Estimate the next weekly occurrence of a given time.

    .. warning:: Both arguments are expected to have the same tzinfo, no internal checks.

    Parameters
    ----------
    now : datetime.datetime
        `datetime.datetime` object of today
    target_time : datetime.time
        Desired `datetime.time`.

    Returns
    -------
    datetime.datetime
        Next `datetime.datetime` object with the desired time.
    """
    target = now.replace(
        second=target_time.second,
        microsecond=target_time.microsecond,
    )
    if (target - now).total_seconds() <= 0:
        return target + dt.timedelta(minutes=1)
    return target


def next_weekday_time_occurrence(
    now: dt.datetime, weekday: Weekday, target_time: dt.time
) -> dt.datetime:
    """
    Estimate the next occurrence of a given weekday and time.

    .. warning:: Arguments `now` and `target_time` are expected to have the same tzinfo,
       no internal checks.

    Parameters
    ----------
    now : datetime.datetime
        `datetime.datetime` object of today
    weekday : Weekday
        Desired :class:`~scheduler.util.Weekday`.
    target_time : datetime.time
        Desired `datetime.time`.

    Returns
    -------
    datetime.datetime
        Next `datetime.datetime` object with the desired weekday and time.
    """
    days = days_to_weekday(now.weekday(), weekday.value)
    if days == 7:
        candidate = next_daily_occurrence(now, target_time)
        if candidate.date() == now.date():
            return candidate

    delta = dt.timedelta(days=days)
    target = now.replace(
        hour=target_time.hour,
        minute=target_time.minute,
        second=target_time.second,
        microsecond=target_time.microsecond,
    )
    return target + delta


JOB_NEXT_DAYLIKE_MAPPING = {
    JobType.MINUTELY: next_minutely_occurrence,
    JobType.HOURLY: next_hourly_occurrence,
    JobType.DAILY: next_daily_occurrence,
}


def are_times_unique(
    timelist: list[dt.time],
) -> bool:
    r"""
    Check if list contains distinct `datetime.time`\ s.

    Parameters
    ----------
    timelist : list[datetime.time]
        List of time objects.

    Returns
    -------
    boolean
        ``True`` if list entries are not equivalent with tzinfo offset.
    """
    ref = dt.datetime(year=1970, month=1, day=1)
    collection = {
        ref.replace(
            hour=time.hour,
            minute=time.minute,
            second=time.second,
            microsecond=time.microsecond,
        )
        + (time.utcoffset() or dt.timedelta())
        for time in timelist
    }
    return len(collection) == len(timelist)


def are_weekday_times_unique(weekday_list: list[Weekday], tzinfo: Optional[dt.tzinfo]) -> bool:
    """
    Check if list contains distinct weekday times.

    .. warning:: Both arguments are expected to be either timezone aware or not
        - no internal checks.

    Parameters
    ----------
    weekday_list : list[Weekday]
        List of weekday objects.

    Returns
    -------
    boolean
        ``True`` if list entries are not equivalent with timezone offset.
    """
    ref = dt.datetime(year=1970, month=1, day=1, tzinfo=tzinfo)
    collection = {
        next_weekday_time_occurrence(ref.astimezone(day.time.tzinfo), day, day.time)
        for day in weekday_list
    }
    return len(collection) == len(weekday_list)



"""
Basic definitions for a abstract `BaseJob` and `BaseScheduler`.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

import datetime as dt
from enum import Enum, auto

from scheduler.base.timingtype import (
    _TimingCyclicList,
    _TimingDailyList,
    _TimingWeeklyList,
)
from scheduler.message import (
    CYCLIC_TYPE_ERROR_MSG,
    DAILY_TYPE_ERROR_MSG,
    HOURLY_TYPE_ERROR_MSG,
    MINUTELY_TYPE_ERROR_MSG,
    WEEKLY_TYPE_ERROR_MSG,
)
from scheduler.trigger import (
    Friday,
    Monday,
    Saturday,
    Sunday,
    Thursday,
    Tuesday,
    Wednesday,
)


class JobType(Enum):
    """Indicate the `JobType` of a |BaseJob|."""

    CYCLIC = auto()
    MINUTELY = auto()
    HOURLY = auto()
    DAILY = auto()
    WEEKLY = auto()


JOB_TYPE_MAPPING = {
    dt.timedelta: JobType.CYCLIC,
    dt.time: JobType.DAILY,
    Monday: JobType.WEEKLY,
    Tuesday: JobType.WEEKLY,
    Wednesday: JobType.WEEKLY,
    Thursday: JobType.WEEKLY,
    Friday: JobType.WEEKLY,
    Saturday: JobType.WEEKLY,
    Sunday: JobType.WEEKLY,
}

JOB_TIMING_TYPE_MAPPING = {
    JobType.CYCLIC: {
        "type": _TimingCyclicList,
        "err": CYCLIC_TYPE_ERROR_MSG,
    },
    JobType.MINUTELY: {
        "type": _TimingDailyList,
        "err": MINUTELY_TYPE_ERROR_MSG,
    },
    JobType.HOURLY: {
        "type": _TimingDailyList,
        "err": HOURLY_TYPE_ERROR_MSG,
    },
    JobType.DAILY: {
        "type": _TimingDailyList,
        "err": DAILY_TYPE_ERROR_MSG,
    },
    JobType.WEEKLY: {
        "type": _TimingWeeklyList,
        "err": WEEKLY_TYPE_ERROR_MSG,
    },
}



"""
Trigger implementations.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

import datetime as dt
from abc import ABC, abstractmethod
from typing import Union


class Weekday(ABC):
    """
    |Weekday| object with time.

    Parameters
    ----------
    time : datetime.time
        Time on the clock at the specific |Weekday|.
    """

    __value: int
    __time: dt.time

    @abstractmethod
    def __init__(self, time: dt.time, value: int) -> None:
        """|Weekday| object with time."""
        self.__time = time
        self.__value = value

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(time={self.time!r})"

    @property
    def time(self) -> dt.time:
        """
        Return time of the |Weekday|.

        Returns
        -------
        datetime.time
            Time on the clock at the specific |Weekday|.
        """
        return self.__time

    @property
    def value(self) -> int:
        """
        Return value of the given |Weekday|.

        Notes
        -----
        Enumeration analogous to datetime library (0: Monday, ... 6: Sunday).

        Returns
        -------
        int
            Value
        """
        return self.__value


# NOTE: pylint missing-class-docstring is just silly here, given functionality and usuage of parent
class Monday(Weekday):  # pylint: disable=missing-class-docstring  # noqa: D101
    __doc__ = Weekday.__doc__

    def __init__(self, time: dt.time = dt.time()) -> None:
        super().__init__(time, 0)


class Tuesday(Weekday):  # pylint: disable=missing-class-docstring  # noqa: D101
    __doc__ = Weekday.__doc__

    def __init__(self, time: dt.time = dt.time()) -> None:
        super().__init__(time, 1)


class Wednesday(Weekday):  # pylint: disable=missing-class-docstring  # noqa: D101
    __doc__ = Weekday.__doc__

    def __init__(self, time: dt.time = dt.time()) -> None:
        super().__init__(time, 2)


class Thursday(Weekday):  # pylint: disable=missing-class-docstring  # noqa: D101
    __doc__ = Weekday.__doc__

    def __init__(self, time: dt.time = dt.time()) -> None:
        super().__init__(time, 3)


class Friday(Weekday):  # pylint: disable=missing-class-docstring  # noqa: D101
    __doc__ = Weekday.__doc__

    def __init__(self, time: dt.time = dt.time()) -> None:
        super().__init__(time, 4)


class Saturday(Weekday):  # pylint: disable=missing-class-docstring  # noqa: D101
    __doc__ = Weekday.__doc__

    def __init__(self, time: dt.time = dt.time()) -> None:
        super().__init__(time, 5)


class Sunday(Weekday):  # pylint: disable=missing-class-docstring  # noqa: D101
    __doc__ = Weekday.__doc__

    def __init__(self, time: dt.time = dt.time()) -> None:
        super().__init__(time, 6)


_Weekday = Union[Monday, Tuesday, Wednesday, Thursday, Friday, Saturday, Sunday]

_weekday_mapping: dict[int, type[_Weekday]] = {
    0: Monday,
    1: Tuesday,
    2: Wednesday,
    3: Thursday,
    4: Friday,
    5: Saturday,
    6: Sunday,
}


def weekday(value: int, time: dt.time = dt.time()) -> Weekday:
    """
    Return |Weekday| from given value with optional time.

    Notes
    -----
    Enumeration analogous to datetime library (0: Monday, ... 6: Sunday).

    Parameters
    ----------
    value : int
        Integer representation of |Weekday|
    time : datetime.time
        Time on the clock at the specific weekday.

    Returns
    -------
    Weekday
        |Weekday| object with given time.
    """
    weekday_cls: type[_Weekday] = _weekday_mapping[value]
    weekday_instance = weekday_cls(time)
    return weekday_instance



"""
Trigger collection.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

from scheduler.trigger.core import (
    Friday,
    Monday,
    Saturday,
    Sunday,
    Thursday,
    Tuesday,
    Wednesday,
    weekday,
)

__all__ = ["Friday", "Monday", "Saturday", "Sunday", "Thursday", "Tuesday", "Wednesday", "weekday"]



"""
Collection of prioritization functions.

For compatibility with the |Scheduler|, the prioritization
functions have to be of type ``Callable[[float, Job, int, int], float]``.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

import random

from scheduler.base.job import BaseJobType
from scheduler.threading.job import Job


def constant_weight_prioritization(
    time_delta: float, job: Job, max_exec: int, job_count: int
) -> float:
    r"""
    Interprets the `Job`'s weight as its priority.

    Return the |Job|'s weight for overdue
    |Job|\ s, otherwise return zero:

    .. math::
        \left(\mathtt{time\_delta},\mathtt{weight}\right)\ {\mapsto}\begin{cases}
        0 & :\ \mathtt{time\_delta}<0\\
        \mathtt{weight} & :\ \mathtt{time\_delta}\geq0
        \end{cases}

    Parameters
    ----------
    time_delta : float
        The time in seconds that a |Job| is overdue.
    job : Job
        The |Job| instance
    max_exec : int
        Limits the number of overdue |Job|\ s that can be executed
        by calling function `Scheduler.exec_jobs()`.
    job_count : int
        Number of scheduled |Job|\ s

    Returns
    -------
    float
        The weight of a |Job| as priority.
    """
    _ = max_exec
    _ = job_count
    if time_delta < 0:
        return 0
    return job.weight


def linear_priority_function(time_delta: float, job: Job, max_exec: int, job_count: int) -> float:
    r"""
    Compute the |Job|\ s default linear priority.

    Linear |Job| prioritization such that the priority increases
    linearly with the amount of time that a |Job| is overdue.
    At the exact time of the scheduled execution, the priority is equal to the
    |Job|\ s weight.

    The function is defined as

    .. math::
        \left(\mathtt{time\_delta},\mathtt{weight}\right)\ {\mapsto}\begin{cases}
        0 & :\ \mathtt{time\_delta}<0\\
        {\left(\mathtt{time\_delta}+1\right)}\cdot\mathtt{weight} & :\ \mathtt{time\_delta}\geq0
        \end{cases}

    Parameters
    ----------
    time_delta : float
        The time in seconds that a |Job| is overdue.
    job : Job
        The |Job| instance
    max_exec : int
        Limits the number of overdue |Job|\ s that can be executed
        by calling function `Scheduler.exec_jobs()`.
    job_count : int
        Number of scheduled |Job|\ s

    Returns
    -------
    float
        The time dependant priority for a |Job|
    """
    _ = max_exec
    _ = job_count

    if time_delta < 0:
        return 0
    return (time_delta + 1) * job.weight


def random_priority_function(time: float, job: Job, max_exec: int, job_count: int) -> float:
    """
    Generate random priority values from weights.

    .. warning:: Not suitable for security relevant purposes.

    The priority generator will return 1 if the random number
    is lower then the |Job|'s weight, otherwise it will return 0.
    """
    _ = time
    _ = max_exec
    _ = job_count
    if random.random() < job.weight:  # nosec
        return 1
    return 0



"""
Implementation of a `asyncio` compatible in-process scheduler.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

from scheduler.asyncio.scheduler import Scheduler
from scheduler.error import SchedulerError

__all__ = ["Scheduler", "SchedulerError"]



"""
Implementation of a `asyncio` compatible in-process scheduler.

Author: Jendrik A. Potyka, Fabian A. Preiss
"""

from __future__ import annotations

import asyncio as aio
import datetime as dt
from asyncio.selector_events import BaseSelectorEventLoop
from collections.abc import Iterable
from logging import Logger
from typing import Any, Callable, Coroutine, Optional

import typeguard as tg

from scheduler.asyncio.job import Job
from scheduler.base.definition import JOB_TYPE_MAPPING, JobType
from scheduler.base.scheduler import BaseScheduler, deprecated, select_jobs_by_tag
from scheduler.base.scheduler_util import check_tzname, create_job_instance, str_cutoff
from scheduler.base.timingtype import (
    TimingCyclic,
    TimingDailyUnion,
    TimingOnceUnion,
    TimingWeeklyUnion,
)
from scheduler.error import SchedulerError
from scheduler.message import (
    CYCLIC_TYPE_ERROR_MSG,
    DAILY_TYPE_ERROR_MSG,
    HOURLY_TYPE_ERROR_MSG,
    MINUTELY_TYPE_ERROR_MSG,
    ONCE_TYPE_ERROR_MSG,
    WEEKLY_TYPE_ERROR_MSG,
)


class Scheduler(BaseScheduler[Job, Callable[..., Coroutine[Any, Any, None]]]):
    r"""
    Implementation of an asyncio scheduler.

    This implementation enables the planning of |AioJob|\ s depending on time
    cycles, fixed times, weekdays, dates, offsets and execution counts.

    Notes
    -----
    Due to the support of `datetime` objects, the |AioScheduler| is able to work
    with timezones.

    Parameters
    ----------
    loop : asyncio.selector_events.BaseSelectorEventLoop
        Set a AsyncIO event loop, default is the global event loop
    tzinfo : datetime.tzinfo
        Set the timezone of the |AioScheduler|.
    logger : Optional[logging.Logger]
        A custom Logger instance.
    """

    def __init__(
        self,
        *,
        loop: Optional[BaseSelectorEventLoop] = None,
        tzinfo: Optional[dt.tzinfo] = None,
        logger: Optional[Logger] = None,
    ):
        super().__init__(logger=logger)
        try:
            self.__loop = loop if loop else aio.get_running_loop()
        except RuntimeError:
            raise SchedulerError("The asyncio Scheduler requires a running event loop.") from None
        self.__tzinfo = tzinfo
        self.__tz_str = check_tzname(tzinfo=tzinfo)

        self._jobs: dict[Job, aio.Task[None]] = {}

    def __repr__(self) -> str:
        return "scheduler.asyncio.scheduler.Scheduler({0}, jobs={{{1}}})".format(
            ", ".join((repr(elem) for elem in (self.__tzinfo,))),
            ", ".join([repr(job) for job in sorted(self.jobs)]),
        )

    def __str__(self) -> str:
        # Scheduler meta heading
        scheduler_headings = "{0}, {1}\n\n".format(*self.__headings())

        # Job table (we join two of the Job._repr() fields into one)
        # columns
        c_align = ("<", "<", "<", "<", ">", ">")
        c_width = (8, 16, 19, 12, 9, 13)
        c_name = (
            "type",
            "function / alias",
            "due at",
            "tzinfo",
            "due in",
            "attempts",
        )
        form = [
            f"{{{idx}:{align}{width}}}" for idx, (align, width) in enumerate(zip(c_align, c_width))
        ]
        if self.__tz_str is None:
            form = form[:3] + form[4:]

        fstring = " ".join(form) + "\n"
        job_table = fstring.format(*c_name)
        job_table += fstring.format(*("-" * width for width in c_width))
        for job in sorted(self.jobs):
            row = job._str()
            entries = (
                row[0],
                str_cutoff(row[1] + row[2], c_width[1], False),
                row[3],
                str_cutoff(row[4] or "", c_width[3], False),
                str_cutoff(row[5], c_width[4], True),
                str_cutoff(f"{row[6]}/{row[7]}", c_width[5], True),
            )
            job_table += fstring.format(*entries)

        return scheduler_headings + job_table

    def __headings(self) -> list[str]:
        headings = [
            f"tzinfo={self.__tz_str}",
            f"#jobs={len(self._jobs)}",
        ]
        return headings

    def __schedule(
        self,
        **kwargs,
    ) -> Job:
        """Encapsulate the `Job` and add the `Scheduler`'s timezone."""
        job: Job = create_job_instance(Job, tzinfo=self.__tzinfo, **kwargs)

        task = self.__loop.create_task(self.__supervise_job(job))
        self._jobs[job] = task

        return job

    async def __supervise_job(self, job: Job) -> None:
        try:
            reference_dt = dt.datetime.now(tz=self.__tzinfo)
            while job.has_attempts_remaining:
                sleep_seconds: float = job.timedelta(reference_dt).total_seconds()
                await aio.sleep(sleep_seconds)

                await job._exec(logger=self._logger)  # pylint: disable=protected-access

                reference_dt = dt.datetime.now(tz=self.__tzinfo)
                job._calc_next_exec(reference_dt)  # pylint: disable=protected-access
        except aio.CancelledError:  # TODO asyncio does not trigger this exception in pytest, why?
            # raised, when `task.cancel()` in `delete_job` was run
            pass  # pragma: no cover
        else:
            self.delete_job(job)

    def delete_job(self, job: Job) -> None:
        """
        Delete a `Job` from the `Scheduler`.

        Parameters
        ----------
        job : Job
            |AioJob| instance to delete.

        Raises
        ------
        SchedulerError
            Raises if the |AioJob| of the argument is not scheduled.
        """
        try:
            task: aio.Task[None] = self._jobs.pop(job)
            _: bool = task.cancel()
        except KeyError:
            raise SchedulerError("An unscheduled Job can not be deleted!") from None

    def delete_jobs(
        self,
        tags: Optional[set[str]] = None,
        any_tag: bool = False,
    ) -> int:
        r"""
        Delete a set of |AioJob|\ s from the |AioScheduler| by tags.

        If no tags or an empty set of tags are given defaults to the deletion
        of all |AioJob|\ s.

        Parameters
        ----------
        tags : Optional[set[str]]
            Set of tags to identify target |AioJob|\ s.
        any_tag : bool
            False: To delete a |AioJob| all tags have to match.
            True: To delete a |AioJob| at least one tag has to match.
        """
        all_jobs: set[Job] = set(self._jobs.keys())
        jobs_to_delete: set[Job]

        if tags is None or tags == set():
            jobs_to_delete = all_jobs
        else:
            jobs_to_delete = select_jobs_by_tag(all_jobs, tags, any_tag)

        for job in jobs_to_delete:
            self.delete_job(job)

        return len(jobs_to_delete)

    def get_jobs(
        self,
        tags: Optional[set[str]] = None,
        any_tag: bool = False,
    ) -> set[Job]:
        r"""
        Get a set of |AioJob|\ s from the |AioScheduler| by tags.

        If no tags or an empty set of tags are given defaults to returning
        all |AioJob|\ s.

        Parameters
        ----------
        tags : set[str]
            Tags to filter scheduled |AioJob|\ s.
            If no tags are given all |AioJob|\ s are returned.
        any_tag : bool
            False: To match a |AioJob| all tags have to match.
            True: To match a |AioJob| at least one tag has to match.

        Returns
        -------
        set[Job]
            Currently scheduled |AioJob|\ s.
        """
        if tags is None or tags == set():
            return self.jobs
        return select_jobs_by_tag(self.jobs, tags, any_tag)

    @deprecated(["delay"])
    def cyclic(
        self, timing: TimingCyclic, handle: Callable[..., Coroutine[Any, Any, None]], **kwargs
    ) -> Job:
        r"""
        Schedule a cyclic `Job`.

        Use a `datetime.timedelta` object or a `list` of `datetime.timedelta` objects
        to schedule a cyclic |AioJob|.

        Parameters
        ----------
        timing : TimingTypeCyclic
            Desired execution time.
        handle : Callable[..., Coroutine[Any, Any, None]]
            Handle to a callback function.

        Returns
        -------
        Job
            Instance of a scheduled |AioJob|.

        Other Parameters
        ----------------
        **kwargs
            |AioJob| properties, optional

            `kwargs` are used to specify |AioJob| properties.

            Here is a list of available |AioJob| properties:

            .. include:: ../_assets/aio_kwargs.rst
        """
        try:
            tg.check_type(timing, TimingCyclic)
        except tg.TypeCheckError as err:
            raise SchedulerError(CYCLIC_TYPE_ERROR_MSG) from err
        return self.__schedule(job_type=JobType.CYCLIC, timing=timing, handle=handle, **kwargs)

    @deprecated(["delay"])
    def minutely(
        self, timing: TimingDailyUnion, handle: Callable[..., Coroutine[Any, Any, None]], **kwargs
    ) -> Job:
        r"""
        Schedule a minutely `Job`.

        Use a `datetime.time` object or a `list` of `datetime.time` objects
        to schedule a |AioJob| every minute.

        Notes
        -----
        If given a `datetime.time` object with a non zero hour or minute property, these
        information will be ignored.

        Parameters
        ----------
        timing : TimingDailyUnion
            Desired execution time(s).
        handle : Callable[..., Coroutine[Any, Any, None]]
            Handle to a callback function.

        Returns
        -------
        Job
            Instance of a scheduled |AioJob|.

        Other Parameters
        ----------------
        **kwargs
            |AioJob| properties, optional

            `kwargs` are used to specify |AioJob| properties.

            Here is a list of available |AioJob| properties:

            .. include:: ../_assets/aio_kwargs.rst
        """
        try:
            tg.check_type(timing, TimingDailyUnion)
        except tg.TypeCheckError as err:
            raise SchedulerError(MINUTELY_TYPE_ERROR_MSG) from err
        return self.__schedule(job_type=JobType.MINUTELY, timing=timing, handle=handle, **kwargs)

    @deprecated(["delay"])
    def hourly(
        self, timing: TimingDailyUnion, handle: Callable[..., Coroutine[Any, Any, None]], **kwargs
    ) -> Job:
        r"""
        Schedule an hourly `Job`.

        Use a `datetime.time` object or a `list` of `datetime.time` objects
        to schedule a |AioJob| every hour.

        Notes
        -----
        If given a `datetime.time` object with a non zero hour property, this information
        will be ignored.

        Parameters
        ----------
        timing : TimingDailyUnion
            Desired execution time(s).
        handle : Callable[..., Coroutine[Any, Any, None]]
            Handle to a callback function.

        Returns
        -------
        Job
            Instance of a scheduled |AioJob|.

        Other Parameters
        ----------------
        **kwargs
            |AioJob| properties, optional

            `kwargs` are used to specify |AioJob| properties.

            Here is a list of available |AioJob| properties:

            .. include:: ../_assets/aio_kwargs.rst
        """
        try:
            tg.check_type(timing, TimingDailyUnion)
        except tg.TypeCheckError as err:
            raise SchedulerError(HOURLY_TYPE_ERROR_MSG) from err
        return self.__schedule(job_type=JobType.HOURLY, timing=timing, handle=handle, **kwargs)

    @deprecated(["delay"])
    def daily(
        self, timing: TimingDailyUnion, handle: Callable[..., Coroutine[Any, Any, None]], **kwargs
    ) -> Job:
        r"""
        Schedule a daily `Job`.

        Use a `datetime.time` object or a `list` of `datetime.time` objects
        to schedule a |AioJob| every day.

        Parameters
        ----------
        timing : TimingDailyUnion
            Desired execution time(s).
        handle : Callable[..., Coroutine[Any, Any, None]]
            Handle to a callback function.

        Returns
        -------
        Job
            Instance of a scheduled |AioJob|.

        Other Parameters
        ----------------
        **kwargs
            |AioJob| properties, optional

            `kwargs` are used to specify |AioJob| properties.

            Here is a list of available |AioJob| properties:

            .. include:: ../_assets/aio_kwargs.rst
        """
        try:
            tg.check_type(timing, TimingDailyUnion)
        except tg.TypeCheckError as err:
            raise SchedulerError(DAILY_TYPE_ERROR_MSG) from err
        return self.__schedule(job_type=JobType.DAILY, timing=timing, handle=handle, **kwargs)

    @deprecated(["delay"])
    def weekly(
        self, timing: TimingWeeklyUnion, handle: Callable[..., Coroutine[Any, Any, None]], **kwargs
    ) -> Job:
        r"""
        Schedule a weekly `Job`.

        Use a `tuple` of a `Weekday` and a `datetime.time` object to define a weekly
        recurring |AioJob|. Combine multiple desired `tuples` in
        a `list`. If the planed execution time is `00:00` the `datetime.time` object
        can be ignored, just pass a `Weekday` without a `tuple`.

        Parameters
        ----------
        timing : TimingWeeklyUnion
            Desired execution time(s).
        handle : Callable[..., Coroutine[Any, Any, None]]
            Handle to a callback function.

        Returns
        -------
        Job
            Instance of a scheduled |AioJob|.

        Other Parameters
        ----------------
        **kwargs
            |AioJob| properties, optional

            `kwargs` are used to specify |AioJob| properties.

            Here is a list of available |AioJob| properties:

            .. include:: ../_assets/aio_kwargs.rst
        """
        try:
            tg.check_type(timing, TimingWeeklyUnion)
        except tg.TypeCheckError as err:
            raise SchedulerError(WEEKLY_TYPE_ERROR_MSG) from err
        return self.__schedule(job_type=JobType.WEEKLY, timing=timing, handle=handle, **kwargs)

    def once(
        self,
        timing: TimingOnceUnion,
        handle: Callable[..., Coroutine[Any, Any, None]],
        *,
        args: Optional[tuple[Any]] = None,
        kwargs: Optional[dict[str, Any]] = None,
        tags: Optional[Iterable[str]] = None,
        alias: Optional[str] = None,
    ) -> Job:
        r"""
        Schedule a oneshot `Job`.

        Parameters
        ----------
        timing : TimingOnceUnion
            Desired execution time.
        handle : Callable[..., Coroutine[Any, Any, None]]
            Handle to a callback function.
        args : tuple[Any]
            Positional argument payload for the function handle within a |AioJob|.
        kwargs : Optional[dict[str, Any]]
            Keyword arguments payload for the function handle within a |AioJob|.
        tags : Optional[Iterable[str]]
            The tags of the |AioJob|.
        alias : Optional[str]
            Overwrites the function handle name in the string representation.

        Returns
        -------
        Job
            Instance of a scheduled |AioJob|.
        """
        try:
            tg.check_type(timing, TimingOnceUnion)
        except tg.TypeCheckError as err:
            raise SchedulerError(ONCE_TYPE_ERROR_MSG) from err
        if isinstance(timing, dt.datetime):
            return self.__schedule(
                job_type=JobType.CYCLIC,
                timing=dt.timedelta(),
                handle=handle,
                args=args,
                kwargs=kwargs,
                max_attempts=1,
                tags=set(tags) if tags else set(),
                alias=alias,
                delay=False,
                start=timing,
            )
        return self.__schedule(
            job_type=JOB_TYPE_MAPPING[type(timing)],
            timing=timing,
            handle=handle,
            args=args,
            kwargs=kwargs,
            max_attempts=1,
            tags=tags,
            alias=alias,
        )

    @property
    def jobs(self) -> set[Job]:
        r"""
        Get the set of all `Job`\ s.

        Returns
        -------
        set[Job]
            Currently scheduled |AioJob|\ s.
        """
        return set(self._jobs.keys())
